from pathlib import Path

import torch.jit

from bioimageio.spec.model.v0_5 import ModelDescr, OnnxWeightsDescr

from .. import __version__
from ..digest_spec import get_member_id, get_test_input_sample
from ..proc_setup import get_pre_and_postprocessing
from ._utils_onnx import get_dynamic_axes


def convert(
    model_descr: ModelDescr,
    output_path: Path,
    *,
    verbose: bool = False,
    opset_version: int = 15,
) -> OnnxWeightsDescr:
    """
    Convert model weights from the PyTorch state_dict format to the ONNX format.

    Args:
        model_descr (Union[v0_4.ModelDescr, v0_5.ModelDescr]):
            The model description object that contains the model and its weights.
        output_path (Path):
            The file path where the ONNX model will be saved.
        verbose (bool, optional):
            If True, will print out detailed information during the ONNX export process. Defaults to False.
        opset_version (int, optional):
            The ONNX opset version to use for the export. Defaults to 15.
    Raises:
        ValueError:
            If the provided model does not have weights in the torchscript format.

    Returns:
        v0_5.OnnxWeightsDescr:
            A descriptor object that contains information about the exported ONNX weights.
    """

    torchscript_descr = model_descr.weights.torchscript
    if torchscript_descr is None:
        raise ValueError(
            "The provided model does not have weights in the torchscript format"
        )

    sample = get_test_input_sample(model_descr)
    procs = get_pre_and_postprocessing(
        model_descr, dataset_for_initial_statistics=[sample]
    )
    procs.pre(sample)
    inputs_numpy = [
        sample.members[get_member_id(ipt)].data.data for ipt in model_descr.inputs
    ]
    inputs_torch = [torch.from_numpy(ipt) for ipt in inputs_numpy]

    weight_reader = torchscript_descr.get_reader()
    model = torch.jit.load(weight_reader)  # type: ignore
    model.to("cpu")
    model = model.eval()  # type: ignore

    with torch.no_grad():
        outputs_original_torch = model(*inputs_torch)  # type: ignore
        if isinstance(outputs_original_torch, torch.Tensor):
            outputs_original_torch = [outputs_original_torch]

        _ = torch.onnx.export(
            model,  # type: ignore
            tuple(inputs_torch),
            str(output_path),
            input_names=[str(d.id) for d in model_descr.inputs],
            output_names=[str(d.id) for d in model_descr.outputs],
            dynamic_axes=get_dynamic_axes(model_descr),
            verbose=verbose,
            opset_version=opset_version,
        )

    return OnnxWeightsDescr(
        source=output_path.absolute(),
        parent="torchscript",
        opset_version=opset_version,
        comment=f"Converted with bioimageio.core {__version__}.",
    )



from typing import (
    Callable,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Union,
)

from typing_extensions import assert_never

from bioimageio.core.digest_spec import get_member_id
from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5

from .proc_ops import (
    AddKnownDatasetStats,
    EnsureDtype,
    Processing,
    UpdateStats,
    get_proc,
)
from .sample import Sample
from .stat_calculators import StatsCalculator
from .stat_measures import (
    DatasetMeasure,
    DatasetMeasureBase,
    Measure,
    MeasureValue,
    SampleMeasure,
    SampleMeasureBase,
)

TensorDescr = Union[
    v0_4.InputTensorDescr,
    v0_4.OutputTensorDescr,
    v0_5.InputTensorDescr,
    v0_5.OutputTensorDescr,
]


class PreAndPostprocessing(NamedTuple):
    pre: List[Processing]
    post: List[Processing]


class _ProcessingCallables(NamedTuple):
    pre: Callable[[Sample], None]
    post: Callable[[Sample], None]


class _SetupProcessing(NamedTuple):
    pre: List[Processing]
    post: List[Processing]
    pre_measures: Set[Measure]
    post_measures: Set[Measure]


class _ApplyProcs:
    def __init__(self, procs: Sequence[Processing]):
        super().__init__()
        self._procs = procs

    def __call__(self, sample: Sample) -> None:
        for op in self._procs:
            op(sample)


def get_pre_and_postprocessing(
    model: AnyModelDescr,
    *,
    dataset_for_initial_statistics: Iterable[Sample],
    keep_updating_initial_dataset_stats: bool = False,
    fixed_dataset_stats: Optional[Mapping[DatasetMeasure, MeasureValue]] = None,
) -> _ProcessingCallables:
    """Creates callables to apply pre- and postprocessing in-place to a sample"""

    setup = setup_pre_and_postprocessing(
        model=model,
        dataset_for_initial_statistics=dataset_for_initial_statistics,
        keep_updating_initial_dataset_stats=keep_updating_initial_dataset_stats,
        fixed_dataset_stats=fixed_dataset_stats,
    )
    return _ProcessingCallables(_ApplyProcs(setup.pre), _ApplyProcs(setup.post))


def setup_pre_and_postprocessing(
    model: AnyModelDescr,
    dataset_for_initial_statistics: Iterable[Sample],
    keep_updating_initial_dataset_stats: bool = False,
    fixed_dataset_stats: Optional[Mapping[DatasetMeasure, MeasureValue]] = None,
) -> PreAndPostprocessing:
    """
    Get pre- and postprocessing operators for a `model` description.
    Used in `bioimageio.core.create_prediction_pipeline"""
    prep, post, prep_meas, post_meas = _prepare_setup_pre_and_postprocessing(model)

    missing_dataset_stats = {
        m
        for m in prep_meas | post_meas
        if fixed_dataset_stats is None or m not in fixed_dataset_stats
    }
    if missing_dataset_stats:
        initial_stats_calc = StatsCalculator(missing_dataset_stats)
        for sample in dataset_for_initial_statistics:
            initial_stats_calc.update(sample)

        initial_stats = initial_stats_calc.finalize()
    else:
        initial_stats = {}

    prep.insert(
        0,
        UpdateStats(
            StatsCalculator(prep_meas, initial_stats),
            keep_updating_initial_dataset_stats=keep_updating_initial_dataset_stats,
        ),
    )
    if post_meas:
        post.insert(
            0,
            UpdateStats(
                StatsCalculator(post_meas, initial_stats),
                keep_updating_initial_dataset_stats=keep_updating_initial_dataset_stats,
            ),
        )

    if fixed_dataset_stats:
        prep.insert(0, AddKnownDatasetStats(fixed_dataset_stats))
        post.insert(0, AddKnownDatasetStats(fixed_dataset_stats))

    return PreAndPostprocessing(prep, post)


class RequiredMeasures(NamedTuple):
    pre: Set[Measure]
    post: Set[Measure]


class RequiredDatasetMeasures(NamedTuple):
    pre: Set[DatasetMeasure]
    post: Set[DatasetMeasure]


class RequiredSampleMeasures(NamedTuple):
    pre: Set[SampleMeasure]
    post: Set[SampleMeasure]


def get_requried_measures(model: AnyModelDescr) -> RequiredMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredMeasures(s.pre_measures, s.post_measures)


def get_required_dataset_measures(model: AnyModelDescr) -> RequiredDatasetMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredDatasetMeasures(
        {m for m in s.pre_measures if isinstance(m, DatasetMeasureBase)},
        {m for m in s.post_measures if isinstance(m, DatasetMeasureBase)},
    )


def get_requried_sample_measures(model: AnyModelDescr) -> RequiredSampleMeasures:
    s = _prepare_setup_pre_and_postprocessing(model)
    return RequiredSampleMeasures(
        {m for m in s.pre_measures if isinstance(m, SampleMeasureBase)},
        {m for m in s.post_measures if isinstance(m, SampleMeasureBase)},
    )


def _prepare_procs(
    tensor_descrs: Union[
        Sequence[v0_4.InputTensorDescr],
        Sequence[v0_5.InputTensorDescr],
        Sequence[v0_4.OutputTensorDescr],
        Sequence[v0_5.OutputTensorDescr],
    ],
) -> List[Processing]:
    procs: List[Processing] = []
    for t_descr in tensor_descrs:
        if isinstance(t_descr, (v0_4.InputTensorDescr, v0_4.OutputTensorDescr)):
            member_id = get_member_id(t_descr)
            procs.append(
                EnsureDtype(input=member_id, output=member_id, dtype=t_descr.data_type)
            )

        if isinstance(t_descr, (v0_4.InputTensorDescr, v0_5.InputTensorDescr)):
            for proc_d in t_descr.preprocessing:
                procs.append(get_proc(proc_d, t_descr))
        elif isinstance(t_descr, (v0_4.OutputTensorDescr, v0_5.OutputTensorDescr)):
            for proc_d in t_descr.postprocessing:
                procs.append(get_proc(proc_d, t_descr))
        else:
            assert_never(t_descr)

        if isinstance(
            t_descr,
            (v0_4.InputTensorDescr, (v0_4.InputTensorDescr, v0_4.OutputTensorDescr)),
        ):
            if len(procs) == 1:
                # remove initial ensure_dtype if there are no other proccessing steps
                assert isinstance(procs[0], EnsureDtype)
                procs = []

            # ensure 0.4 models get float32 input
            # which has been the implicit assumption for 0.4
            member_id = get_member_id(t_descr)
            procs.append(
                EnsureDtype(input=member_id, output=member_id, dtype="float32")
            )

    return procs


def _prepare_setup_pre_and_postprocessing(model: AnyModelDescr) -> _SetupProcessing:
    if isinstance(model, v0_4.ModelDescr):
        pre = _prepare_procs(model.inputs)
        post = _prepare_procs(model.outputs)
    elif isinstance(model, v0_5.ModelDescr):
        pre = _prepare_procs(model.inputs)
        post = _prepare_procs(model.outputs)
    else:
        assert_never(model)

    return _SetupProcessing(
        pre=pre,
        post=post,
        pre_measures={m for proc in pre for m in proc.required_measures},
        post_measures={m for proc in post for m in proc.required_measures},
    )



from pathlib import Path
from typing import Literal, Optional

import pytest

from bioimageio.core import load_model
from bioimageio.core.commands import package, validate_format
from bioimageio.core.commands import test as command_tst
from bioimageio.spec import AnyModelDescr


@pytest.fixture(scope="module")
def model(unet2d_nuclei_broad_model: str):
    return load_model(unet2d_nuclei_broad_model, perform_io_checks=False)


@pytest.mark.parametrize(
    "weight_format",
    [
        "all",
        "pytorch_state_dict",
    ],
)
def test_package(
    weight_format: Literal["all", "pytorch_state_dict"],
    model: AnyModelDescr,
    tmp_path: Path,
):
    assert package(model, weight_format=weight_format, path=tmp_path / "out.zip") == 0


def test_validate_format(model: AnyModelDescr):
    assert validate_format(model) == 0


@pytest.mark.parametrize(
    "weight_format,devices", [("all", None), ("pytorch_state_dict", "cpu")]
)
def test_test(
    weight_format: Literal["all", "pytorch_state_dict"],
    devices: Optional[str],
    model: AnyModelDescr,
):
    assert command_tst(model, weight_format=weight_format, devices=devices) == 0



from difflib import HtmlDiff, unified_diff
from typing import Sequence

from typing_extensions import Literal, assert_never


def compare(
    a: Sequence[str],
    b: Sequence[str],
    name_a: str = "source",
    name_b: str = "updated",
    *,
    diff_format: Literal["unified", "html"],
):
    if diff_format == "html":
        diff = HtmlDiff().make_file(a, b, name_a, name_b, charset="utf-8")
    elif diff_format == "unified":
        diff = "\n".join(
            unified_diff(
                a,
                b,
                name_a,
                name_b,
                lineterm="",
            )
        )
    else:
        assert_never(diff_format)

    return diff



import collections.abc
from abc import ABC, abstractmethod
from dataclasses import InitVar, dataclass, field
from typing import (
    Collection,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import numpy as np
import scipy  # pyright: ignore[reportMissingTypeStubs]
import xarray as xr
from typing_extensions import Self, assert_never

from bioimageio.core.digest_spec import get_member_id
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import (
    _convert_proc,  # pyright: ignore [reportPrivateUsage]
)

from ._op_base import BlockedOperator, Operator
from .axis import AxisId, PerAxis
from .block import Block
from .common import DTypeStr, MemberId
from .sample import Sample, SampleBlock, SampleBlockWithOrigin
from .stat_calculators import StatsCalculator
from .stat_measures import (
    DatasetMean,
    DatasetMeasure,
    DatasetPercentile,
    DatasetStd,
    MeanMeasure,
    Measure,
    MeasureValue,
    SampleMean,
    SampleQuantile,
    SampleStd,
    Stat,
    StdMeasure,
)
from .tensor import Tensor


def _convert_axis_ids(
    axes: v0_4.AxesInCZYX,
    mode: Literal["per_sample", "per_dataset"],
) -> Tuple[AxisId, ...]:
    if not isinstance(axes, str):
        return tuple(axes)

    if mode == "per_sample":
        ret = []
    elif mode == "per_dataset":
        ret = [v0_5.BATCH_AXIS_ID]
    else:
        assert_never(mode)

    ret.extend([AxisId(a) for a in axes])
    return tuple(ret)


@dataclass
class _SimpleOperator(BlockedOperator, ABC):
    input: MemberId
    output: MemberId

    @property
    def required_measures(self) -> Collection[Measure]:
        return set()

    @abstractmethod
    def get_output_shape(self, input_shape: PerAxis[int]) -> PerAxis[int]: ...

    def __call__(self, sample: Union[Sample, SampleBlock]) -> None:
        if self.input not in sample.members:
            return

        input_tensor = sample.members[self.input]
        output_tensor = self._apply(input_tensor, sample.stat)

        if self.output in sample.members:
            assert (
                sample.members[self.output].tagged_shape == output_tensor.tagged_shape
            )

        if isinstance(sample, Sample):
            sample.members[self.output] = output_tensor
        elif isinstance(sample, SampleBlock):
            b = sample.blocks[self.input]
            sample.blocks[self.output] = Block(
                sample_shape=self.get_output_shape(sample.shape[self.input]),
                data=output_tensor,
                inner_slice=b.inner_slice,
                halo=b.halo,
                block_index=b.block_index,
                blocks_in_sample=b.blocks_in_sample,
            )
        else:
            assert_never(sample)

    @abstractmethod
    def _apply(self, x: Tensor, stat: Stat) -> Tensor: ...


@dataclass
class AddKnownDatasetStats(BlockedOperator):
    dataset_stats: Mapping[DatasetMeasure, MeasureValue]

    @property
    def required_measures(self) -> Set[Measure]:
        return set()

    def __call__(self, sample: Union[Sample, SampleBlock]) -> None:
        sample.stat.update(self.dataset_stats.items())


# @dataclass
# class UpdateStats(Operator):
#     """Calculates sample and/or dataset measures"""

#     measures: Union[Sequence[Measure], Set[Measure], Mapping[Measure, MeasureValue]]
#     """sample and dataset `measuers` to be calculated by this operator. Initial/fixed
#     dataset measure values may be given, see `keep_updating_dataset_stats` for details.
#     """
#     keep_updating_dataset_stats: Optional[bool] = None
#     """indicates if operator calls should keep updating dataset statistics or not

#     default (None): if `measures` is a `Mapping` (i.e. initial measure values are
#     given) no further updates to dataset statistics is conducted, otherwise (w.o.
#     initial measure values) dataset statistics are updated by each processed sample.
#     """
#     _keep_updating_dataset_stats: bool = field(init=False)
#     _stats_calculator: StatsCalculator = field(init=False)

#     @property
#     def required_measures(self) -> Set[Measure]:
#         return set()

#     def __post_init__(self):
#         self._stats_calculator = StatsCalculator(self.measures)
#         if self.keep_updating_dataset_stats is None:
#             self._keep_updating_dataset_stats = not isinstance(self.measures, collections.abc.Mapping)
#         else:
#             self._keep_updating_dataset_stats = self.keep_updating_dataset_stats

#     def __call__(self, sample_block: SampleBlockWithOrigin> None:
#         if self._keep_updating_dataset_stats:
#             sample.stat.update(self._stats_calculator.update_and_get_all(sample))
#         else:
#             sample.stat.update(self._stats_calculator.skip_update_and_get_all(sample))


@dataclass
class UpdateStats(Operator):
    """Calculates sample and/or dataset measures"""

    stats_calculator: StatsCalculator
    """`StatsCalculator` to be used by this operator."""
    keep_updating_initial_dataset_stats: bool = False
    """indicates if operator calls should keep updating initial dataset statistics or not;
    if the `stats_calculator` was not provided with any initial dataset statistics,
    these are always updated with every new sample.
    """
    _keep_updating_dataset_stats: bool = field(init=False)

    @property
    def required_measures(self) -> Set[Measure]:
        return set()

    def __post_init__(self):
        self._keep_updating_dataset_stats = (
            self.keep_updating_initial_dataset_stats
            or not self.stats_calculator.has_dataset_measures
        )

    def __call__(self, sample: Union[Sample, SampleBlockWithOrigin]) -> None:
        if isinstance(sample, SampleBlockWithOrigin):
            # update stats with whole sample on first block
            if sample.block_index != 0:
                return

            origin = sample.origin
        else:
            origin = sample

        if self._keep_updating_dataset_stats:
            sample.stat.update(self.stats_calculator.update_and_get_all(origin))
        else:
            sample.stat.update(self.stats_calculator.skip_update_and_get_all(origin))


@dataclass
class Binarize(_SimpleOperator):
    """'output = tensor > threshold'."""

    threshold: Union[float, Sequence[float]]
    axis: Optional[AxisId] = None

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return x > self.threshold

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls, descr: Union[v0_4.BinarizeDescr, v0_5.BinarizeDescr], member_id: MemberId
    ) -> Self:
        if isinstance(descr.kwargs, (v0_4.BinarizeKwargs, v0_5.BinarizeKwargs)):
            return cls(
                input=member_id, output=member_id, threshold=descr.kwargs.threshold
            )
        elif isinstance(descr.kwargs, v0_5.BinarizeAlongAxisKwargs):
            return cls(
                input=member_id,
                output=member_id,
                threshold=descr.kwargs.threshold,
                axis=descr.kwargs.axis,
            )
        else:
            assert_never(descr.kwargs)


@dataclass
class Clip(_SimpleOperator):
    min: Optional[float] = None
    """minimum value for clipping"""
    max: Optional[float] = None
    """maximum value for clipping"""

    def __post_init__(self):
        assert self.min is not None or self.max is not None, "missing min or max value"
        assert (
            self.min is None or self.max is None or self.min < self.max
        ), f"expected min < max, but {self.min} !< {self.max}"

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return x.clip(self.min, self.max)

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls, descr: Union[v0_4.ClipDescr, v0_5.ClipDescr], member_id: MemberId
    ) -> Self:
        return cls(
            input=member_id,
            output=member_id,
            min=descr.kwargs.min,
            max=descr.kwargs.max,
        )


@dataclass
class EnsureDtype(_SimpleOperator):
    dtype: DTypeStr

    @classmethod
    def from_proc_descr(cls, descr: v0_5.EnsureDtypeDescr, member_id: MemberId):
        return cls(input=member_id, output=member_id, dtype=descr.kwargs.dtype)

    def get_descr(self):
        return v0_5.EnsureDtypeDescr(kwargs=v0_5.EnsureDtypeKwargs(dtype=self.dtype))

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return x.astype(self.dtype)


@dataclass
class ScaleLinear(_SimpleOperator):
    gain: Union[float, xr.DataArray] = 1.0
    """multiplicative factor"""

    offset: Union[float, xr.DataArray] = 0.0
    """additive term"""

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return x * self.gain + self.offset

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls,
        descr: Union[v0_4.ScaleLinearDescr, v0_5.ScaleLinearDescr],
        member_id: MemberId,
    ) -> Self:
        kwargs = descr.kwargs
        if isinstance(kwargs, v0_5.ScaleLinearKwargs):
            axis = None
        elif isinstance(kwargs, v0_5.ScaleLinearAlongAxisKwargs):
            axis = kwargs.axis
        elif isinstance(kwargs, v0_4.ScaleLinearKwargs):
            if kwargs.axes is not None:
                raise NotImplementedError(
                    "model.v0_4.ScaleLinearKwargs with axes not implemented, please consider updating the model to v0_5."
                )
            axis = None
        else:
            assert_never(kwargs)

        if axis:
            gain = xr.DataArray(np.atleast_1d(kwargs.gain), dims=axis)
            offset = xr.DataArray(np.atleast_1d(kwargs.offset), dims=axis)
        else:
            assert (
                isinstance(kwargs.gain, (float, int)) or len(kwargs.gain) == 1
            ), kwargs.gain
            gain = (
                kwargs.gain if isinstance(kwargs.gain, (float, int)) else kwargs.gain[0]
            )
            assert isinstance(kwargs.offset, (float, int)) or len(kwargs.offset) == 1
            offset = (
                kwargs.offset
                if isinstance(kwargs.offset, (float, int))
                else kwargs.offset[0]
            )

        return cls(input=member_id, output=member_id, gain=gain, offset=offset)


@dataclass
class ScaleMeanVariance(_SimpleOperator):
    axes: Optional[Sequence[AxisId]] = None
    reference_tensor: Optional[MemberId] = None
    eps: float = 1e-6
    mean: Union[SampleMean, DatasetMean] = field(init=False)
    std: Union[SampleStd, DatasetStd] = field(init=False)
    ref_mean: Union[SampleMean, DatasetMean] = field(init=False)
    ref_std: Union[SampleStd, DatasetStd] = field(init=False)

    @property
    def required_measures(self):
        return {self.mean, self.std, self.ref_mean, self.ref_std}

    def __post_init__(self):
        axes = None if self.axes is None else tuple(self.axes)
        ref_tensor = self.reference_tensor or self.input
        if axes is None or AxisId("batch") not in axes:
            Mean = SampleMean
            Std = SampleStd
        else:
            Mean = DatasetMean
            Std = DatasetStd

        self.mean = Mean(member_id=self.input, axes=axes)
        self.std = Std(member_id=self.input, axes=axes)
        self.ref_mean = Mean(member_id=ref_tensor, axes=axes)
        self.ref_std = Std(member_id=ref_tensor, axes=axes)

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        mean = stat[self.mean]
        std = stat[self.std] + self.eps
        ref_mean = stat[self.ref_mean]
        ref_std = stat[self.ref_std] + self.eps
        return (x - mean) / std * ref_std + ref_mean

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls,
        descr: Union[v0_4.ScaleMeanVarianceDescr, v0_5.ScaleMeanVarianceDescr],
        member_id: MemberId,
    ) -> Self:
        kwargs = descr.kwargs
        _, axes = _get_axes(descr.kwargs)

        return cls(
            input=member_id,
            output=member_id,
            reference_tensor=MemberId(str(kwargs.reference_tensor)),
            axes=axes,
            eps=kwargs.eps,
        )


def _get_axes(
    kwargs: Union[
        v0_4.ZeroMeanUnitVarianceKwargs,
        v0_5.ZeroMeanUnitVarianceKwargs,
        v0_4.ScaleRangeKwargs,
        v0_5.ScaleRangeKwargs,
        v0_4.ScaleMeanVarianceKwargs,
        v0_5.ScaleMeanVarianceKwargs,
    ],
) -> Tuple[bool, Optional[Tuple[AxisId, ...]]]:
    if kwargs.axes is None:
        return True, None
    elif isinstance(kwargs.axes, str):
        axes = _convert_axis_ids(kwargs.axes, kwargs["mode"])
        return AxisId("b") in axes, axes
    elif isinstance(kwargs.axes, collections.abc.Sequence):
        axes = tuple(kwargs.axes)
        return AxisId("batch") in axes, axes
    else:
        assert_never(kwargs.axes)


@dataclass
class ScaleRange(_SimpleOperator):
    lower_percentile: InitVar[Optional[Union[SampleQuantile, DatasetPercentile]]] = None
    upper_percentile: InitVar[Optional[Union[SampleQuantile, DatasetPercentile]]] = None
    lower: Union[SampleQuantile, DatasetPercentile] = field(init=False)
    upper: Union[SampleQuantile, DatasetPercentile] = field(init=False)

    eps: float = 1e-6

    def __post_init__(
        self,
        lower_percentile: Optional[Union[SampleQuantile, DatasetPercentile]],
        upper_percentile: Optional[Union[SampleQuantile, DatasetPercentile]],
    ):
        if lower_percentile is None:
            tid = self.input if upper_percentile is None else upper_percentile.member_id
            self.lower = DatasetPercentile(q=0.0, member_id=tid)
        else:
            self.lower = lower_percentile

        if upper_percentile is None:
            self.upper = DatasetPercentile(q=1.0, member_id=self.lower.member_id)
        else:
            self.upper = upper_percentile

        assert self.lower.member_id == self.upper.member_id
        assert self.lower.q < self.upper.q
        assert self.lower.axes == self.upper.axes

    @property
    def required_measures(self):
        return {self.lower, self.upper}

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls,
        descr: Union[v0_4.ScaleRangeDescr, v0_5.ScaleRangeDescr],
        member_id: MemberId,
    ):
        kwargs = descr.kwargs
        ref_tensor = (
            member_id
            if kwargs.reference_tensor is None
            else MemberId(str(kwargs.reference_tensor))
        )
        dataset_mode, axes = _get_axes(descr.kwargs)
        if dataset_mode:
            Percentile = DatasetPercentile
        else:
            Percentile = SampleQuantile

        return cls(
            input=member_id,
            output=member_id,
            lower_percentile=Percentile(
                q=kwargs.min_percentile / 100, axes=axes, member_id=ref_tensor
            ),
            upper_percentile=Percentile(
                q=kwargs.max_percentile / 100, axes=axes, member_id=ref_tensor
            ),
        )

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        lower = stat[self.lower]
        upper = stat[self.upper]
        return (x - lower) / (upper - lower + self.eps)

    def get_descr(self):
        assert self.lower.axes == self.upper.axes
        assert self.lower.member_id == self.upper.member_id

        return v0_5.ScaleRangeDescr(
            kwargs=v0_5.ScaleRangeKwargs(
                axes=self.lower.axes,
                min_percentile=self.lower.q * 100,
                max_percentile=self.upper.q * 100,
                eps=self.eps,
                reference_tensor=self.lower.member_id,
            )
        )


@dataclass
class Sigmoid(_SimpleOperator):
    """1 / (1 + e^(-input))."""

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return Tensor(1.0 / (1.0 + np.exp(-x)), dims=x.dims)

    @property
    def required_measures(self) -> Collection[Measure]:
        return {}

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls, descr: Union[v0_4.SigmoidDescr, v0_5.SigmoidDescr], member_id: MemberId
    ) -> Self:
        assert isinstance(descr, (v0_4.SigmoidDescr, v0_5.SigmoidDescr))
        return cls(input=member_id, output=member_id)

    def get_descr(self):
        return v0_5.SigmoidDescr()


@dataclass
class Softmax(_SimpleOperator):
    """Softmax activation function."""

    axis: AxisId = AxisId("channel")

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        axis_idx = x.dims.index(self.axis)
        result = scipy.special.softmax(x.data, axis=axis_idx)
        result_xr = xr.DataArray(result, dims=x.dims)
        return Tensor.from_xarray(result_xr)

    @property
    def required_measures(self) -> Collection[Measure]:
        return {}

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(cls, descr: v0_5.SoftmaxDescr, member_id: MemberId) -> Self:
        assert isinstance(descr, v0_5.SoftmaxDescr)
        return cls(input=member_id, output=member_id, axis=descr.kwargs.axis)

    def get_descr(self):
        return v0_5.SoftmaxDescr(kwargs=v0_5.SoftmaxKwargs(axis=self.axis))


@dataclass
class ZeroMeanUnitVariance(_SimpleOperator):
    """normalize to zero mean, unit variance."""

    mean: MeanMeasure
    std: StdMeasure

    eps: float = 1e-6

    def __post_init__(self):
        assert self.mean.axes == self.std.axes

    @property
    def required_measures(self) -> Set[Union[MeanMeasure, StdMeasure]]:
        return {self.mean, self.std}

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls,
        descr: Union[v0_4.ZeroMeanUnitVarianceDescr, v0_5.ZeroMeanUnitVarianceDescr],
        member_id: MemberId,
    ):
        dataset_mode, axes = _get_axes(descr.kwargs)

        if dataset_mode:
            Mean = DatasetMean
            Std = DatasetStd
        else:
            Mean = SampleMean
            Std = SampleStd

        return cls(
            input=member_id,
            output=member_id,
            mean=Mean(axes=axes, member_id=member_id),
            std=Std(axes=axes, member_id=member_id),
        )

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        mean = stat[self.mean]
        std = stat[self.std]
        return (x - mean) / (std + self.eps)

    def get_descr(self):
        return v0_5.ZeroMeanUnitVarianceDescr(
            kwargs=v0_5.ZeroMeanUnitVarianceKwargs(axes=self.mean.axes, eps=self.eps)
        )


@dataclass
class FixedZeroMeanUnitVariance(_SimpleOperator):
    """normalize to zero mean, unit variance with precomputed values."""

    mean: Union[float, xr.DataArray]
    std: Union[float, xr.DataArray]

    eps: float = 1e-6

    def __post_init__(self):
        assert (
            isinstance(self.mean, (int, float))
            or isinstance(self.std, (int, float))
            or self.mean.dims == self.std.dims
        )

    def get_output_shape(
        self, input_shape: Mapping[AxisId, int]
    ) -> Mapping[AxisId, int]:
        return input_shape

    @classmethod
    def from_proc_descr(
        cls,
        descr: v0_5.FixedZeroMeanUnitVarianceDescr,
        member_id: MemberId,
    ) -> Self:
        if isinstance(descr.kwargs, v0_5.FixedZeroMeanUnitVarianceKwargs):
            dims = None
        elif isinstance(descr.kwargs, v0_5.FixedZeroMeanUnitVarianceAlongAxisKwargs):
            dims = (AxisId(descr.kwargs.axis),)
        else:
            assert_never(descr.kwargs)

        return cls(
            input=member_id,
            output=member_id,
            mean=xr.DataArray(descr.kwargs.mean, dims=dims),
            std=xr.DataArray(descr.kwargs.std, dims=dims),
        )

    def get_descr(self):
        if isinstance(self.mean, (int, float)):
            assert isinstance(self.std, (int, float))
            kwargs = v0_5.FixedZeroMeanUnitVarianceKwargs(mean=self.mean, std=self.std)
        else:
            assert isinstance(self.std, xr.DataArray)
            assert len(self.mean.dims) == 1
            kwargs = v0_5.FixedZeroMeanUnitVarianceAlongAxisKwargs(
                axis=AxisId(str(self.mean.dims[0])),
                mean=list(self.mean),
                std=list(self.std),
            )

        return v0_5.FixedZeroMeanUnitVarianceDescr(kwargs=kwargs)

    def _apply(self, x: Tensor, stat: Stat) -> Tensor:
        return (x - self.mean) / (self.std + self.eps)


ProcDescr = Union[
    v0_4.PreprocessingDescr,
    v0_4.PostprocessingDescr,
    v0_5.PreprocessingDescr,
    v0_5.PostprocessingDescr,
]

Processing = Union[
    AddKnownDatasetStats,
    Binarize,
    Clip,
    EnsureDtype,
    FixedZeroMeanUnitVariance,
    ScaleLinear,
    ScaleMeanVariance,
    ScaleRange,
    Sigmoid,
    Softmax,
    UpdateStats,
    ZeroMeanUnitVariance,
]


def get_proc(
    proc_descr: ProcDescr,
    tensor_descr: Union[
        v0_4.InputTensorDescr,
        v0_4.OutputTensorDescr,
        v0_5.InputTensorDescr,
        v0_5.OutputTensorDescr,
    ],
) -> Processing:
    member_id = get_member_id(tensor_descr)

    if isinstance(proc_descr, (v0_4.BinarizeDescr, v0_5.BinarizeDescr)):
        return Binarize.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, (v0_4.ClipDescr, v0_5.ClipDescr)):
        return Clip.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, v0_5.EnsureDtypeDescr):
        return EnsureDtype.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, v0_5.FixedZeroMeanUnitVarianceDescr):
        return FixedZeroMeanUnitVariance.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, (v0_4.ScaleLinearDescr, v0_5.ScaleLinearDescr)):
        return ScaleLinear.from_proc_descr(proc_descr, member_id)
    elif isinstance(
        proc_descr, (v0_4.ScaleMeanVarianceDescr, v0_5.ScaleMeanVarianceDescr)
    ):
        return ScaleMeanVariance.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, (v0_4.ScaleRangeDescr, v0_5.ScaleRangeDescr)):
        return ScaleRange.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, (v0_4.SigmoidDescr, v0_5.SigmoidDescr)):
        return Sigmoid.from_proc_descr(proc_descr, member_id)
    elif (
        isinstance(proc_descr, v0_4.ZeroMeanUnitVarianceDescr)
        and proc_descr.kwargs.mode == "fixed"
    ):
        if not isinstance(
            tensor_descr, (v0_4.InputTensorDescr, v0_4.OutputTensorDescr)
        ):
            raise TypeError(
                "Expected v0_4 tensor description for v0_4 processing description"
            )

        v5_proc_descr = _convert_proc(proc_descr, tensor_descr.axes)
        assert isinstance(v5_proc_descr, v0_5.FixedZeroMeanUnitVarianceDescr)
        return FixedZeroMeanUnitVariance.from_proc_descr(v5_proc_descr, member_id)
    elif isinstance(
        proc_descr,
        (v0_4.ZeroMeanUnitVarianceDescr, v0_5.ZeroMeanUnitVarianceDescr),
    ):
        return ZeroMeanUnitVariance.from_proc_descr(proc_descr, member_id)
    elif isinstance(proc_descr, v0_5.SoftmaxDescr):
        return Softmax.from_proc_descr(proc_descr, member_id)
    else:
        assert_never(proc_descr)



def test_save_bioimageio_package(unet2d_nuclei_broad_model: str):
    from bioimageio.spec._package import save_bioimageio_package

    _ = save_bioimageio_package(
        unet2d_nuclei_broad_model,
        weights_priority_order=("pytorch_state_dict",),
    )



import itertools
from dataclasses import dataclass
from functools import cached_property
from math import floor, prod
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from loguru import logger
from typing_extensions import Self

from .axis import AxisId, PerAxis
from .common import (
    BlockIndex,
    Frozen,
    Halo,
    HaloLike,
    MemberId,
    PadWidth,
    PerMember,
    SliceInfo,
    TotalNumberOfBlocks,
)


@dataclass
class LinearAxisTransform:
    axis: AxisId
    scale: float
    offset: int

    def compute(self, s: int, round: Callable[[float], int] = floor) -> int:
        return round(s * self.scale) + self.offset


@dataclass(frozen=True)
class BlockMeta:
    """Block meta data of a sample member (a tensor in a sample)

    Figure for illustration:
    The first 2d block (dashed) of a sample member (**bold**).
    The inner slice (thin) is expanded by a halo in both dimensions on both sides.
    The outer slice reaches from the sample member origin (0, 0) to the right halo point.

    ```terminal
    ┌ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─  ─ ─ ─ ─ ─ ─ ─ ┐
    ╷ halo(left)                         ╷
    ╷                                    ╷
    ╷  (0, 0)┏━━━━━━━━━━━━━━━━━┯━━━━━━━━━┯━━━➔
    ╷        ┃                 │         ╷  sample member
    ╷        ┃      inner      │         ╷
    ╷        ┃   (and outer)   │  outer  ╷
    ╷        ┃      slice      │  slice  ╷
    ╷        ┃                 │         ╷
    ╷        ┣─────────────────┘         ╷
    ╷        ┃   outer slice             ╷
    ╷        ┃               halo(right) ╷
    └ ─ ─ ─ ─┃─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─┘
             ⬇
    ```

    note:
    - Inner and outer slices are specified in sample member coordinates.
    - The outer_slice of a block at the sample edge may overlap by more than the
        halo with the neighboring block (the inner slices will not overlap though).

    """

    sample_shape: PerAxis[int]
    """the axis sizes of the whole (unblocked) sample"""

    inner_slice: PerAxis[SliceInfo]
    """inner region (without halo) wrt the sample"""

    halo: PerAxis[Halo]
    """halo enlarging the inner region to the block's sizes"""

    block_index: BlockIndex
    """the i-th block of the sample"""

    blocks_in_sample: TotalNumberOfBlocks
    """total number of blocks in the sample"""

    @cached_property
    def shape(self) -> PerAxis[int]:
        """axis lengths of the block"""
        return Frozen(
            {
                a: s.stop - s.start + (sum(self.halo[a]) if a in self.halo else 0)
                for a, s in self.inner_slice.items()
            }
        )

    @cached_property
    def padding(self) -> PerAxis[PadWidth]:
        """padding to realize the halo at the sample edge
        where we cannot simply enlarge the inner slice"""
        return Frozen(
            {
                a: PadWidth(
                    (
                        self.halo[a].left
                        - (self.inner_slice[a].start - self.outer_slice[a].start)
                        if a in self.halo
                        else 0
                    ),
                    (
                        self.halo[a].right
                        - (self.outer_slice[a].stop - self.inner_slice[a].stop)
                        if a in self.halo
                        else 0
                    ),
                )
                for a in self.inner_slice
            }
        )

    @cached_property
    def outer_slice(self) -> PerAxis[SliceInfo]:
        """slice of the outer block (without padding) wrt the sample"""
        return Frozen(
            {
                a: SliceInfo(
                    max(
                        0,
                        min(
                            self.inner_slice[a].start
                            - (self.halo[a].left if a in self.halo else 0),
                            self.sample_shape[a]
                            - self.inner_shape[a]
                            - (self.halo[a].left if a in self.halo else 0),
                        ),
                    ),
                    min(
                        self.sample_shape[a],
                        self.inner_slice[a].stop
                        + (self.halo[a].right if a in self.halo else 0),
                    ),
                )
                for a in self.inner_slice
            }
        )

    @cached_property
    def inner_shape(self) -> PerAxis[int]:
        """axis lengths of the inner region (without halo)"""
        return Frozen({a: s.stop - s.start for a, s in self.inner_slice.items()})

    @cached_property
    def local_slice(self) -> PerAxis[SliceInfo]:
        """inner slice wrt the block, **not** the sample"""
        return Frozen(
            {
                a: SliceInfo(
                    self.halo[a].left,
                    self.halo[a].left + self.inner_shape[a],
                )
                for a in self.inner_slice
            }
        )

    @property
    def dims(self) -> Collection[AxisId]:
        return set(self.inner_shape)

    @property
    def tagged_shape(self) -> PerAxis[int]:
        """alias for shape"""
        return self.shape

    @property
    def inner_slice_wo_overlap(self):
        """subslice of the inner slice, such that all `inner_slice_wo_overlap` can be
        stiched together trivially to form the original sample.

        This can also be used to calculate statistics
        without overrepresenting block edge regions."""
        # TODO: update inner_slice_wo_overlap when adding block overlap
        return self.inner_slice

    def __post_init__(self):
        # freeze mutable inputs
        if not isinstance(self.sample_shape, Frozen):
            object.__setattr__(self, "sample_shape", Frozen(self.sample_shape))

        if not isinstance(self.inner_slice, Frozen):
            object.__setattr__(self, "inner_slice", Frozen(self.inner_slice))

        if not isinstance(self.halo, Frozen):
            object.__setattr__(self, "halo", Frozen(self.halo))

        assert all(
            a in self.sample_shape for a in self.inner_slice
        ), "block has axes not present in sample"

        assert all(
            a in self.inner_slice for a in self.halo
        ), "halo has axes not present in block"

        if any(s > self.sample_shape[a] for a, s in self.shape.items()):
            logger.warning(
                "block {} larger than sample {}", self.shape, self.sample_shape
            )

    def get_transformed(
        self, new_axes: PerAxis[Union[LinearAxisTransform, int]]
    ) -> Self:
        return self.__class__(
            sample_shape={
                a: (
                    trf
                    if isinstance(trf, int)
                    else trf.compute(self.sample_shape[trf.axis])
                )
                for a, trf in new_axes.items()
            },
            inner_slice={
                a: (
                    SliceInfo(0, trf)
                    if isinstance(trf, int)
                    else SliceInfo(
                        trf.compute(self.inner_slice[trf.axis].start),
                        trf.compute(self.inner_slice[trf.axis].stop),
                    )
                )
                for a, trf in new_axes.items()
            },
            halo={
                a: (
                    Halo(0, 0)
                    if isinstance(trf, int)
                    else Halo(self.halo[trf.axis].left, self.halo[trf.axis].right)
                )
                for a, trf in new_axes.items()
            },
            block_index=self.block_index,
            blocks_in_sample=self.blocks_in_sample,
        )


def split_shape_into_blocks(
    shape: PerAxis[int],
    block_shape: PerAxis[int],
    halo: PerAxis[HaloLike],
    stride: Optional[PerAxis[int]] = None,
) -> Tuple[TotalNumberOfBlocks, Generator[BlockMeta, Any, None]]:
    assert all(a in shape for a in block_shape), (
        tuple(shape),
        set(block_shape),
    )
    if any(shape[a] < block_shape[a] for a in block_shape):
        # TODO: allow larger blockshape
        raise ValueError(f"shape {shape} is smaller than block shape {block_shape}")

    assert all(a in shape for a in halo), (tuple(shape), set(halo))

    # fill in default halo (0) and block axis length (from tensor shape)
    halo = {a: Halo.create(halo.get(a, 0)) for a in shape}
    block_shape = {a: block_shape.get(a, s) for a, s in shape.items()}
    if stride is None:
        stride = {}

    inner_1d_slices: Dict[AxisId, List[SliceInfo]] = {}
    for a, s in shape.items():
        inner_size = block_shape[a] - sum(halo[a])
        stride_1d = stride.get(a, inner_size)
        inner_1d_slices[a] = [
            SliceInfo(min(p, s - inner_size), min(p + inner_size, s))
            for p in range(0, s, stride_1d)
        ]

    n_blocks = prod(map(len, inner_1d_slices.values()))

    return n_blocks, _block_meta_generator(
        shape,
        blocks_in_sample=n_blocks,
        inner_1d_slices=inner_1d_slices,
        halo=halo,
    )


def _block_meta_generator(
    sample_shape: PerAxis[int],
    *,
    blocks_in_sample: int,
    inner_1d_slices: Dict[AxisId, List[SliceInfo]],
    halo: PerAxis[HaloLike],
):
    assert all(a in sample_shape for a in halo)

    halo = {a: Halo.create(halo.get(a, 0)) for a in inner_1d_slices}
    for i, nd_tile in enumerate(itertools.product(*inner_1d_slices.values())):
        inner_slice: PerAxis[SliceInfo] = dict(zip(inner_1d_slices, nd_tile))

        yield BlockMeta(
            sample_shape=sample_shape,
            inner_slice=inner_slice,
            halo=halo,
            block_index=i,
            blocks_in_sample=blocks_in_sample,
        )


def split_multiple_shapes_into_blocks(
    shapes: PerMember[PerAxis[int]],
    block_shapes: PerMember[PerAxis[int]],
    *,
    halo: PerMember[PerAxis[HaloLike]],
    strides: Optional[PerMember[PerAxis[int]]] = None,
    broadcast: bool = False,
) -> Tuple[TotalNumberOfBlocks, Iterable[PerMember[BlockMeta]]]:
    if unknown_blocks := [t for t in block_shapes if t not in shapes]:
        raise ValueError(
            f"block shape specified for unknown tensors: {unknown_blocks}."
        )

    if not block_shapes:
        block_shapes = shapes

    if not broadcast and (
        missing_blocks := [t for t in shapes if t not in block_shapes]
    ):
        raise ValueError(
            f"no block shape specified for {missing_blocks}."
            + " Set `broadcast` to True if these tensors should be repeated"
            + " as a whole for each block."
        )

    if extra_halo := [t for t in halo if t not in block_shapes]:
        raise ValueError(
            f"`halo` specified for tensors without block shape: {extra_halo}."
        )

    if strides is None:
        strides = {}

    assert not (
        unknown_block := [t for t in strides if t not in block_shapes]
    ), f"`stride` specified for tensors without block shape: {unknown_block}"

    blocks: Dict[MemberId, Iterable[BlockMeta]] = {}
    n_blocks: Dict[MemberId, TotalNumberOfBlocks] = {}
    for t in block_shapes:
        n_blocks[t], blocks[t] = split_shape_into_blocks(
            shape=shapes[t],
            block_shape=block_shapes[t],
            halo=halo.get(t, {}),
            stride=strides.get(t),
        )
        assert n_blocks[t] > 0, n_blocks

    assert len(blocks) > 0, blocks
    assert len(n_blocks) > 0, n_blocks
    unique_n_blocks = set(n_blocks.values())
    n = max(unique_n_blocks)
    if len(unique_n_blocks) == 2 and 1 in unique_n_blocks:
        if not broadcast:
            raise ValueError(
                "Mismatch for total number of blocks due to unsplit (single block)"
                + f" tensors: {n_blocks}. Set `broadcast` to True if you want to"
                + " repeat unsplit (single block) tensors."
            )

        blocks = {
            t: _repeat_single_block(block_gen, n) if n_blocks[t] == 1 else block_gen
            for t, block_gen in blocks.items()
        }
    elif len(unique_n_blocks) != 1:
        raise ValueError(f"Mismatch for total number of blocks: {n_blocks}")

    return n, _aligned_blocks_generator(n, blocks)


def _aligned_blocks_generator(
    n: TotalNumberOfBlocks, blocks: Dict[MemberId, Iterable[BlockMeta]]
):
    iterators = {t: iter(gen) for t, gen in blocks.items()}
    for _ in range(n):
        yield {t: next(it) for t, it in iterators.items()}


def _repeat_single_block(block_generator: Iterable[BlockMeta], n: TotalNumberOfBlocks):
    round_two = False
    for block in block_generator:
        assert not round_two
        for _ in range(n):
            yield block

        round_two = True



"""bioimageio CLI

Note: Some docstrings use a hair space ' '
      to place the added '(default: ...)' on a new line.
"""

import json
import shutil
import subprocess
import sys
from abc import ABC
from argparse import RawTextHelpFormatter
from difflib import SequenceMatcher
from functools import cached_property
from io import StringIO
from pathlib import Path
from pprint import pformat, pprint
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import rich.markdown
from loguru import logger
from pydantic import AliasChoices, BaseModel, Field, model_validator
from pydantic_settings import (
    BaseSettings,
    CliPositionalArg,
    CliSettingsSource,
    CliSubCommand,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from tqdm import tqdm
from typing_extensions import assert_never

import bioimageio.spec
from bioimageio.core import __version__
from bioimageio.spec import (
    AnyModelDescr,
    InvalidDescr,
    ResourceDescr,
    load_description,
    save_bioimageio_yaml_only,
    settings,
    update_format,
    update_hashes,
)
from bioimageio.spec._internal.io import is_yaml_value
from bioimageio.spec._internal.io_utils import open_bioimageio_yaml
from bioimageio.spec._internal.types import FormatVersionPlaceholder, NotEmpty
from bioimageio.spec.dataset import DatasetDescr
from bioimageio.spec.model import ModelDescr, v0_4, v0_5
from bioimageio.spec.notebook import NotebookDescr
from bioimageio.spec.utils import ensure_description_is_model, get_reader, write_yaml

from .commands import WeightFormatArgAll, WeightFormatArgAny, package, test
from .common import MemberId, SampleId, SupportedWeightsFormat
from .digest_spec import get_member_ids, load_sample_for_model
from .io import load_dataset_stat, save_dataset_stat, save_sample
from .prediction import create_prediction_pipeline
from .proc_setup import (
    DatasetMeasure,
    Measure,
    MeasureValue,
    StatsCalculator,
    get_required_dataset_measures,
)
from .sample import Sample
from .stat_measures import Stat
from .utils import compare
from .weight_converters._add_weights import add_weights

WEIGHT_FORMAT_ALIASES = AliasChoices(
    "weight-format",
    "weights-format",
    "weight_format",
    "weights_format",
)


class CmdBase(BaseModel, use_attribute_docstrings=True, cli_implicit_flags=True):
    pass


class ArgMixin(BaseModel, use_attribute_docstrings=True, cli_implicit_flags=True):
    pass


class WithSummaryLogging(ArgMixin):
    summary: List[Union[Literal["display"], Path]] = Field(
        default_factory=lambda: ["display"],
        examples=[
            Path("summary.md"),
            Path("bioimageio_summaries/"),
            ["display", Path("summary.md")],
        ],
    )
    """Display the validation summary or save it as JSON, Markdown or HTML.
    The format is chosen based on the suffix: `.json`, `.md`, `.html`.
    If a folder is given (path w/o suffix) the summary is saved in all formats.
    Choose/add `"display"` to render the validation summary to the terminal.
    """

    def log(self, descr: Union[ResourceDescr, InvalidDescr]):
        _ = descr.validation_summary.log(self.summary)


class WithSource(ArgMixin):
    source: CliPositionalArg[str]
    """Url/path to a (folder with a) bioimageio.yaml/rdf.yaml file
    or a bioimage.io resource identifier, e.g. 'affable-shark'"""

    @cached_property
    def descr(self):
        return load_description(self.source)

    @property
    def descr_id(self) -> str:
        """a more user-friendly description id
        (replacing legacy ids with their nicknames)
        """
        if isinstance(self.descr, InvalidDescr):
            return str(getattr(self.descr, "id", getattr(self.descr, "name")))

        nickname = None
        if (
            isinstance(self.descr.config, v0_5.Config)
            and (bio_config := self.descr.config.bioimageio)
            and bio_config.model_extra is not None
        ):
            nickname = bio_config.model_extra.get("nickname")

        return str(nickname or self.descr.id or self.descr.name)


class ValidateFormatCmd(CmdBase, WithSource, WithSummaryLogging):
    """Validate the meta data format of a bioimageio resource."""

    perform_io_checks: bool = Field(
        settings.perform_io_checks, alias="perform-io-checks"
    )
    """Wether or not to perform validations that requires downloading remote files.
    Note: Default value is set by `BIOIMAGEIO_PERFORM_IO_CHECKS` environment variable.
    """

    @cached_property
    def descr(self):
        return load_description(self.source, perform_io_checks=self.perform_io_checks)

    def run(self):
        self.log(self.descr)
        sys.exit(
            0
            if self.descr.validation_summary.status in ("valid-format", "passed")
            else 1
        )


class TestCmd(CmdBase, WithSource, WithSummaryLogging):
    """Test a bioimageio resource (beyond meta data formatting)."""

    weight_format: WeightFormatArgAll = Field(
        "all",
        alias="weight-format",
        validation_alias=WEIGHT_FORMAT_ALIASES,
    )
    """The weight format to limit testing to.

    (only relevant for model resources)"""

    devices: Optional[List[str]] = None
    """Device(s) to use for testing"""

    runtime_env: Union[Literal["currently-active", "as-described"], Path] = Field(
        "currently-active", alias="runtime-env"
    )
    """The python environment to run the tests in
        - `"currently-active"`: use active Python interpreter
        - `"as-described"`: generate a conda environment YAML file based on the model
            weights description.
        - A path to a conda environment YAML.
          Note: The `bioimageio.core` dependency will be added automatically if not present.
    """

    determinism: Literal["seed_only", "full"] = "seed_only"
    """Modes to improve reproducibility of test outputs."""

    stop_early: bool = Field(
        False, alias="stop-early", validation_alias=AliasChoices("stop-early", "x")
    )
    """Do not run further subtests after a failed one."""

    format_version: Union[FormatVersionPlaceholder, str] = Field(
        "discover", alias="format-version"
    )
    """The format version to use for testing.
        - 'latest': Use the latest implemented format version for the given resource type (may trigger auto updating)
        - 'discover': Use the format version as described in the resource description
        - '0.4', '0.5', ...: Use the specified format version (may trigger auto updating)
    """

    def run(self):
        sys.exit(
            test(
                self.descr,
                weight_format=self.weight_format,
                devices=self.devices,
                summary=self.summary,
                runtime_env=self.runtime_env,
                determinism=self.determinism,
                format_version=self.format_version,
            )
        )


class PackageCmd(CmdBase, WithSource, WithSummaryLogging):
    """Save a resource's metadata with its associated files."""

    path: CliPositionalArg[Path]
    """The path to write the (zipped) package to.
    If it does not have a `.zip` suffix
    this command will save the package as an unzipped folder instead."""

    weight_format: WeightFormatArgAll = Field(
        "all",
        alias="weight-format",
        validation_alias=WEIGHT_FORMAT_ALIASES,
    )
    """The weight format to include in the package (for model descriptions only)."""

    def run(self):
        if isinstance(self.descr, InvalidDescr):
            self.log(self.descr)
            raise ValueError(f"Invalid {self.descr.type} description.")

        sys.exit(
            package(
                self.descr,
                self.path,
                weight_format=self.weight_format,
            )
        )


def _get_stat(
    model_descr: AnyModelDescr,
    dataset: Iterable[Sample],
    dataset_length: int,
    stats_path: Path,
) -> Mapping[DatasetMeasure, MeasureValue]:
    req_dataset_meas, _ = get_required_dataset_measures(model_descr)
    if not req_dataset_meas:
        return {}

    req_dataset_meas, _ = get_required_dataset_measures(model_descr)

    if stats_path.exists():
        logger.info("loading precomputed dataset measures from {}", stats_path)
        stat = load_dataset_stat(stats_path)
        for m in req_dataset_meas:
            if m not in stat:
                raise ValueError(f"Missing {m} in {stats_path}")

        return stat

    stats_calc = StatsCalculator(req_dataset_meas)

    for sample in tqdm(
        dataset, total=dataset_length, desc="precomputing dataset stats", unit="sample"
    ):
        stats_calc.update(sample)

    stat = stats_calc.finalize()
    save_dataset_stat(stat, stats_path)

    return stat


class UpdateCmdBase(CmdBase, WithSource, ABC):
    output: Union[Literal["display", "stdout"], Path] = "display"
    """Output updated bioimageio.yaml to the terminal or write to a file.
    Notes:
    - `"display"`: Render to the terminal with syntax highlighting.
    - `"stdout"`: Write to sys.stdout without syntax highligthing.
      (More convenient for copying the updated bioimageio.yaml from the terminal.)
    """

    diff: Union[bool, Path] = Field(True, alias="diff")
    """Output a diff of original and updated bioimageio.yaml.
    If a given path has an `.html` extension, a standalone HTML file is written,
    otherwise the diff is saved in unified diff format (pure text).
    """

    exclude_unset: bool = Field(True, alias="exclude-unset")
    """Exclude fields that have not explicitly be set."""

    exclude_defaults: bool = Field(False, alias="exclude-defaults")
    """Exclude fields that have the default value (even if set explicitly)."""

    @cached_property
    def updated(self) -> Union[ResourceDescr, InvalidDescr]:
        raise NotImplementedError

    def run(self):
        original_yaml = open_bioimageio_yaml(self.source).unparsed_content
        assert isinstance(original_yaml, str)
        stream = StringIO()

        save_bioimageio_yaml_only(
            self.updated,
            stream,
            exclude_unset=self.exclude_unset,
            exclude_defaults=self.exclude_defaults,
        )
        updated_yaml = stream.getvalue()

        diff = compare(
            original_yaml.split("\n"),
            updated_yaml.split("\n"),
            diff_format=(
                "html"
                if isinstance(self.diff, Path) and self.diff.suffix == ".html"
                else "unified"
            ),
        )

        if isinstance(self.diff, Path):
            _ = self.diff.write_text(diff, encoding="utf-8")
        elif self.diff:
            console = rich.console.Console()
            diff_md = f"## Diff\n\n````````diff\n{diff}\n````````"
            console.print(rich.markdown.Markdown(diff_md))

        if isinstance(self.output, Path):
            _ = self.output.write_text(updated_yaml, encoding="utf-8")
            logger.info(f"written updated description to {self.output}")
        elif self.output == "display":
            updated_md = f"## Updated bioimageio.yaml\n\n```yaml\n{updated_yaml}\n```"
            rich.console.Console().print(rich.markdown.Markdown(updated_md))
        elif self.output == "stdout":
            print(updated_yaml)
        else:
            assert_never(self.output)

        if isinstance(self.updated, InvalidDescr):
            logger.warning("Update resulted in invalid description")
            _ = self.updated.validation_summary.display()


class UpdateFormatCmd(UpdateCmdBase):
    """Update the metadata format to the latest format version."""

    exclude_defaults: bool = Field(True, alias="exclude-defaults")
    """Exclude fields that have the default value (even if set explicitly).

    Note:
        The update process sets most unset fields explicitly with their default value.
    """

    perform_io_checks: bool = Field(
        settings.perform_io_checks, alias="perform-io-checks"
    )
    """Wether or not to attempt validation that may require file download.
    If `True` file hash values are added if not present."""

    @cached_property
    def updated(self):
        return update_format(
            self.source,
            exclude_defaults=self.exclude_defaults,
            perform_io_checks=self.perform_io_checks,
        )


class UpdateHashesCmd(UpdateCmdBase):
    """Create a bioimageio.yaml description with updated file hashes."""

    @cached_property
    def updated(self):
        return update_hashes(self.source)


class PredictCmd(CmdBase, WithSource):
    """Run inference on your data with a bioimage.io model."""

    inputs: NotEmpty[List[Union[str, NotEmpty[List[str]]]]] = Field(
        default_factory=lambda: ["{input_id}/001.tif"]
    )
    """Model input sample paths (for each input tensor)

    The input paths are expected to have shape...
     - (n_samples,) or (n_samples,1) for models expecting a single input tensor
     - (n_samples,) containing the substring '{input_id}', or
     - (n_samples, n_model_inputs) to provide each input tensor path explicitly.

    All substrings that are replaced by metadata from the model description:
    - '{model_id}'
    - '{input_id}'

    Example inputs to process sample 'a' and 'b'
    for a model expecting a 'raw' and a 'mask' input tensor:
    --inputs="[[\\"a_raw.tif\\",\\"a_mask.tif\\"],[\\"b_raw.tif\\",\\"b_mask.tif\\"]]"
    (Note that JSON double quotes need to be escaped.)

    Alternatively a `bioimageio-cli.yaml` (or `bioimageio-cli.json`) file
    may provide the arguments, e.g.:
    ```yaml
    inputs:
    - [a_raw.tif, a_mask.tif]
    - [b_raw.tif, b_mask.tif]
    ```

    `.npy` and any file extension supported by imageio are supported.
     Aavailable formats are listed at
    https://imageio.readthedocs.io/en/stable/formats/index.html#all-formats.
    Some formats have additional dependencies.

     
    """

    outputs: Union[str, NotEmpty[Tuple[str, ...]]] = (
        "outputs_{model_id}/{output_id}/{sample_id}.tif"
    )
    """Model output path pattern (per output tensor)

    All substrings that are replaced:
    - '{model_id}' (from model description)
    - '{output_id}' (from model description)
    - '{sample_id}' (extracted from input paths)

     
    """

    overwrite: bool = False
    """allow overwriting existing output files"""

    blockwise: bool = False
    """process inputs blockwise"""

    stats: Path = Path("dataset_statistics.json")
    """path to dataset statistics
    (will be written if it does not exist,
    but the model requires statistical dataset measures)
     """

    preview: bool = False
    """preview which files would be processed
    and what outputs would be generated."""

    weight_format: WeightFormatArgAny = Field(
        "any",
        alias="weight-format",
        validation_alias=WEIGHT_FORMAT_ALIASES,
    )
    """The weight format to use."""

    example: bool = False
    """generate and run an example

    1. downloads example model inputs
    2. creates a `{model_id}_example` folder
    3. writes input arguments to `{model_id}_example/bioimageio-cli.yaml`
    4. executes a preview dry-run
    5. executes prediction with example input

     
    """

    def _example(self):
        model_descr = ensure_description_is_model(self.descr)
        input_ids = get_member_ids(model_descr.inputs)
        example_inputs = (
            model_descr.sample_inputs
            if isinstance(model_descr, v0_4.ModelDescr)
            else [
                t
                for ipt in model_descr.inputs
                if (t := ipt.sample_tensor or ipt.test_tensor)
            ]
        )
        if not example_inputs:
            raise ValueError(f"{self.descr_id} does not specify any example inputs.")

        inputs001: List[str] = []
        example_path = Path(f"{self.descr_id}_example")
        example_path.mkdir(exist_ok=True)

        for t, src in zip(input_ids, example_inputs):
            reader = get_reader(src)
            dst = Path(f"{example_path}/{t}/001{reader.suffix}")
            dst.parent.mkdir(parents=True, exist_ok=True)
            inputs001.append(dst.as_posix())
            with dst.open("wb") as f:
                shutil.copyfileobj(reader, f)

        inputs = [inputs001]
        output_pattern = f"{example_path}/outputs/{{output_id}}/{{sample_id}}.tif"

        bioimageio_cli_path = example_path / YAML_FILE
        stats_file = "dataset_statistics.json"
        stats = (example_path / stats_file).as_posix()
        cli_example_args = dict(
            inputs=inputs,
            outputs=output_pattern,
            stats=stats_file,
            blockwise=self.blockwise,
        )
        assert is_yaml_value(cli_example_args), cli_example_args
        write_yaml(
            cli_example_args,
            bioimageio_cli_path,
        )

        yaml_file_content = None

        # escaped double quotes
        inputs_json = json.dumps(inputs)
        inputs_escaped = inputs_json.replace('"', r"\"")
        source_escaped = self.source.replace('"', r"\"")

        def get_example_command(preview: bool, escape: bool = False):
            q: str = '"' if escape else ""

            return [
                "bioimageio",
                "predict",
                # --no-preview not supported for py=3.8
                *(["--preview"] if preview else []),
                "--overwrite",
                *(["--blockwise"] if self.blockwise else []),
                f"--stats={q}{stats}{q}",
                f"--inputs={q}{inputs_escaped if escape else inputs_json}{q}",
                f"--outputs={q}{output_pattern}{q}",
                f"{q}{source_escaped if escape else self.source}{q}",
            ]

        if Path(YAML_FILE).exists():
            logger.info(
                "temporarily removing '{}' to execute example prediction", YAML_FILE
            )
            yaml_file_content = Path(YAML_FILE).read_bytes()
            Path(YAML_FILE).unlink()

        try:
            _ = subprocess.run(get_example_command(True), check=True)
            _ = subprocess.run(get_example_command(False), check=True)
        finally:
            if yaml_file_content is not None:
                _ = Path(YAML_FILE).write_bytes(yaml_file_content)
                logger.debug("restored '{}'", YAML_FILE)

        print(
            "🎉 Sucessfully ran example prediction!\n"
            + "To predict the example input using the CLI example config file"
            + f" {example_path / YAML_FILE}, execute `bioimageio predict` from {example_path}:\n"
            + f"$ cd {str(example_path)}\n"
            + f'$ bioimageio predict "{source_escaped}"\n\n'
            + "Alternatively run the following command"
            + " in the current workind directory, not the example folder:\n$ "
            + " ".join(get_example_command(False, escape=True))
            + f"\n(note that a local '{JSON_FILE}' or '{YAML_FILE}' may interfere with this)"
        )

    def run(self):
        if self.example:
            return self._example()

        model_descr = ensure_description_is_model(self.descr)

        input_ids = get_member_ids(model_descr.inputs)
        output_ids = get_member_ids(model_descr.outputs)

        minimum_input_ids = tuple(
            str(ipt.id) if isinstance(ipt, v0_5.InputTensorDescr) else str(ipt.name)
            for ipt in model_descr.inputs
            if not isinstance(ipt, v0_5.InputTensorDescr) or not ipt.optional
        )
        maximum_input_ids = tuple(
            str(ipt.id) if isinstance(ipt, v0_5.InputTensorDescr) else str(ipt.name)
            for ipt in model_descr.inputs
        )

        def expand_inputs(i: int, ipt: Union[str, Sequence[str]]) -> Tuple[str, ...]:
            if isinstance(ipt, str):
                ipts = tuple(
                    ipt.format(model_id=self.descr_id, input_id=t) for t in input_ids
                )
            else:
                ipts = tuple(
                    p.format(model_id=self.descr_id, input_id=t)
                    for t, p in zip(input_ids, ipt)
                )

            if len(set(ipts)) < len(ipts):
                if len(minimum_input_ids) == len(maximum_input_ids):
                    n = len(minimum_input_ids)
                else:
                    n = f"{len(minimum_input_ids)}-{len(maximum_input_ids)}"

                raise ValueError(
                    f"[input sample #{i}] Include '{{input_id}}' in path pattern or explicitly specify {n} distinct input paths (got {ipt})"
                )

            if len(ipts) < len(minimum_input_ids):
                raise ValueError(
                    f"[input sample #{i}] Expected at least {len(minimum_input_ids)} inputs {minimum_input_ids}, got {ipts}"
                )

            if len(ipts) > len(maximum_input_ids):
                raise ValueError(
                    f"Expected at most {len(maximum_input_ids)} inputs {maximum_input_ids}, got {ipts}"
                )

            return ipts

        inputs = [expand_inputs(i, ipt) for i, ipt in enumerate(self.inputs, start=1)]

        sample_paths_in = [
            {t: Path(p) for t, p in zip(input_ids, ipts)} for ipts in inputs
        ]

        sample_ids = _get_sample_ids(sample_paths_in)

        def expand_outputs():
            if isinstance(self.outputs, str):
                outputs = [
                    tuple(
                        Path(
                            self.outputs.format(
                                model_id=self.descr_id, output_id=t, sample_id=s
                            )
                        )
                        for t in output_ids
                    )
                    for s in sample_ids
                ]
            else:
                outputs = [
                    tuple(
                        Path(p.format(model_id=self.descr_id, output_id=t, sample_id=s))
                        for t, p in zip(output_ids, self.outputs)
                    )
                    for s in sample_ids
                ]

            for i, out in enumerate(outputs, start=1):
                if len(set(out)) < len(out):
                    raise ValueError(
                        f"[output sample #{i}] Include '{{output_id}}' in path pattern or explicitly specify {len(output_ids)} distinct output paths (got {out})"
                    )

                if len(out) != len(output_ids):
                    raise ValueError(
                        f"[output sample #{i}] Expected {len(output_ids)} outputs {output_ids}, got {out}"
                    )

            return outputs

        outputs = expand_outputs()

        sample_paths_out = [
            {MemberId(t): Path(p) for t, p in zip(output_ids, out)} for out in outputs
        ]

        if not self.overwrite:
            for sample_paths in sample_paths_out:
                for p in sample_paths.values():
                    if p.exists():
                        raise FileExistsError(
                            f"{p} already exists. use --overwrite to (re-)write outputs anyway."
                        )
        if self.preview:
            print("🛈 bioimageio prediction preview structure:")
            pprint(
                {
                    "{sample_id}": dict(
                        inputs={"{input_id}": "<input path>"},
                        outputs={"{output_id}": "<output path>"},
                    )
                }
            )
            print("🔎 bioimageio prediction preview output:")
            pprint(
                {
                    s: dict(
                        inputs={t: p.as_posix() for t, p in sp_in.items()},
                        outputs={t: p.as_posix() for t, p in sp_out.items()},
                    )
                    for s, sp_in, sp_out in zip(
                        sample_ids, sample_paths_in, sample_paths_out
                    )
                }
            )
            return

        def input_dataset(stat: Stat):
            for s, sp_in in zip(sample_ids, sample_paths_in):
                yield load_sample_for_model(
                    model=model_descr,
                    paths=sp_in,
                    stat=stat,
                    sample_id=s,
                )

        stat: Dict[Measure, MeasureValue] = dict(
            _get_stat(
                model_descr, input_dataset({}), len(sample_ids), self.stats
            ).items()
        )

        pp = create_prediction_pipeline(
            model_descr,
            weight_format=None if self.weight_format == "any" else self.weight_format,
        )
        predict_method = (
            pp.predict_sample_with_blocking
            if self.blockwise
            else pp.predict_sample_without_blocking
        )

        for sample_in, sp_out in tqdm(
            zip(input_dataset(dict(stat)), sample_paths_out),
            total=len(inputs),
            desc=f"predict with {self.descr_id}",
            unit="sample",
        ):
            sample_out = predict_method(sample_in)
            save_sample(sp_out, sample_out)


class AddWeightsCmd(CmdBase, WithSource, WithSummaryLogging):
    output: CliPositionalArg[Path]
    """The path to write the updated model package to."""

    source_format: Optional[SupportedWeightsFormat] = Field(None, alias="source-format")
    """Exclusively use these weights to convert to other formats."""

    target_format: Optional[SupportedWeightsFormat] = Field(None, alias="target-format")
    """Exclusively add this weight format."""

    verbose: bool = False
    """Log more (error) output."""

    tracing: bool = True
    """Allow tracing when converting pytorch_state_dict to torchscript
    (still uses scripting if possible)."""

    def run(self):
        model_descr = ensure_description_is_model(self.descr)
        if isinstance(model_descr, v0_4.ModelDescr):
            raise TypeError(
                f"model format {model_descr.format_version} not supported."
                + " Please update the model first."
            )
        updated_model_descr = add_weights(
            model_descr,
            output_path=self.output,
            source_format=self.source_format,
            target_format=self.target_format,
            verbose=self.verbose,
            allow_tracing=self.tracing,
        )
        self.log(updated_model_descr)


JSON_FILE = "bioimageio-cli.json"
YAML_FILE = "bioimageio-cli.yaml"


class Bioimageio(
    BaseSettings,
    cli_implicit_flags=True,
    cli_parse_args=True,
    cli_prog_name="bioimageio",
    cli_use_class_docs_for_groups=True,
    use_attribute_docstrings=True,
):
    """bioimageio - CLI for bioimage.io resources 🦒"""

    model_config = SettingsConfigDict(
        json_file=JSON_FILE,
        yaml_file=YAML_FILE,
    )

    validate_format: CliSubCommand[ValidateFormatCmd] = Field(alias="validate-format")
    "Check a resource's metadata format"

    test: CliSubCommand[TestCmd]
    "Test a bioimageio resource (beyond meta data formatting)"

    package: CliSubCommand[PackageCmd]
    "Package a resource"

    predict: CliSubCommand[PredictCmd]
    "Predict with a model resource"

    update_format: CliSubCommand[UpdateFormatCmd] = Field(alias="update-format")
    """Update the metadata format"""

    update_hashes: CliSubCommand[UpdateHashesCmd] = Field(alias="update-hashes")
    """Create a bioimageio.yaml description with updated file hashes."""

    add_weights: CliSubCommand[AddWeightsCmd] = Field(alias="add-weights")
    """Add additional weights to the model descriptions converted from available
    formats to improve deployability."""

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        cli: CliSettingsSource[BaseSettings] = CliSettingsSource(
            settings_cls,
            cli_parse_args=True,
            formatter_class=RawTextHelpFormatter,
        )
        sys_args = pformat(sys.argv)
        logger.info("starting CLI with arguments:\n{}", sys_args)
        return (
            cli,
            init_settings,
            YamlConfigSettingsSource(settings_cls),
            JsonConfigSettingsSource(settings_cls),
        )

    @model_validator(mode="before")
    @classmethod
    def _log(cls, data: Any):
        logger.info(
            "loaded CLI input:\n{}",
            pformat({k: v for k, v in data.items() if v is not None}),
        )
        return data

    def run(self):
        logger.info(
            "executing CLI command:\n{}",
            pformat({k: v for k, v in self.model_dump().items() if v is not None}),
        )
        cmd = (
            self.add_weights
            or self.package
            or self.predict
            or self.test
            or self.update_format
            or self.update_hashes
            or self.validate_format
        )
        assert cmd is not None
        cmd.run()


assert isinstance(Bioimageio.__doc__, str)
Bioimageio.__doc__ += f"""

library versions:
  bioimageio.core {__version__}
  bioimageio.spec {bioimageio.spec.__version__}

spec format versions:
        model RDF {ModelDescr.implemented_format_version}
      dataset RDF {DatasetDescr.implemented_format_version}
     notebook RDF {NotebookDescr.implemented_format_version}

"""


def _get_sample_ids(
    input_paths: Sequence[Mapping[MemberId, Path]],
) -> Sequence[SampleId]:
    """Get sample ids for given input paths, based on the common path per sample.

    Falls back to sample01, samle02, etc..."""

    matcher = SequenceMatcher()

    def get_common_seq(seqs: Sequence[Sequence[str]]) -> Sequence[str]:
        """extract a common sequence from multiple sequences
        (order sensitive; strips whitespace and slashes)
        """
        common = seqs[0]

        for seq in seqs[1:]:
            if not seq:
                continue
            matcher.set_seqs(common, seq)
            i, _, size = matcher.find_longest_match()
            common = common[i : i + size]

        if isinstance(common, str):
            common = common.strip().strip("/")
        else:
            common = [cs for c in common if (cs := c.strip().strip("/"))]

        if not common:
            raise ValueError(f"failed to find common sequence for {seqs}")

        return common

    def get_shorter_diff(seqs: Sequence[Sequence[str]]) -> List[Sequence[str]]:
        """get a shorter sequence whose entries are still unique
        (order sensitive, not minimal sequence)
        """
        min_seq_len = min(len(s) for s in seqs)
        # cut from the start
        for start in range(min_seq_len - 1, -1, -1):
            shortened = [s[start:] for s in seqs]
            if len(set(shortened)) == len(seqs):
                min_seq_len -= start
                break
        else:
            seen: Set[Sequence[str]] = set()
            dupes = [s for s in seqs if s in seen or seen.add(s)]
            raise ValueError(f"Found duplicate entries {dupes}")

        # cut from the end
        for end in range(min_seq_len - 1, 1, -1):
            shortened = [s[:end] for s in shortened]
            if len(set(shortened)) == len(seqs):
                break

        return shortened

    full_tensor_ids = [
        sorted(
            p.resolve().with_suffix("").as_posix() for p in input_sample_paths.values()
        )
        for input_sample_paths in input_paths
    ]
    try:
        long_sample_ids = [get_common_seq(t) for t in full_tensor_ids]
        sample_ids = get_shorter_diff(long_sample_ids)
    except ValueError as e:
        raise ValueError(f"failed to extract sample ids: {e}")

    return sample_ids



from dataclasses import dataclass
from typing import (
    Any,
    Generator,
    Iterable,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import Self

from .axis import PerAxis
from .block_meta import BlockMeta, LinearAxisTransform, split_shape_into_blocks
from .common import (
    Halo,
    HaloLike,
    PadMode,
    SliceInfo,
    TotalNumberOfBlocks,
)
from .tensor import Tensor


@dataclass(frozen=True)
class Block(BlockMeta):
    """A block/tile of a (larger) tensor"""

    data: Tensor
    """the block's tensor, e.g. a (padded) slice of some larger, original tensor"""

    @property
    def inner_data(self):
        return self.data[self.local_slice]

    def __post_init__(self):
        super().__post_init__()
        assert not any(v == -1 for v in self.sample_shape.values()), self.sample_shape
        for a, s in self.data.sizes.items():
            slice_ = self.inner_slice[a]
            halo = self.halo.get(a, Halo(0, 0))
            assert s == halo.left + (slice_.stop - slice_.start) + halo.right, (
                s,
                slice_,
                halo,
            )

    @classmethod
    def from_sample_member(
        cls,
        sample_member: Tensor,
        block: BlockMeta,
        *,
        pad_mode: PadMode,
    ) -> Self:
        return cls(
            data=sample_member[block.outer_slice].pad(block.padding, pad_mode),
            sample_shape=sample_member.tagged_shape,
            inner_slice=block.inner_slice,
            halo=block.halo,
            block_index=block.block_index,
            blocks_in_sample=block.blocks_in_sample,
        )

    def get_transformed(
        self, new_axes: PerAxis[Union[LinearAxisTransform, int]]
    ) -> Self:
        raise NotImplementedError

    @classmethod
    def from_meta(cls, meta: BlockMeta, data: Tensor) -> Self:
        return cls(
            sample_shape={
                k: data.tagged_shape[k] if v == -1 else v
                for k, v in meta.sample_shape.items()
            },
            inner_slice={
                k: (
                    SliceInfo(start=v.start, stop=data.tagged_shape[k])
                    if v.stop == -1
                    else v
                )
                for k, v in meta.inner_slice.items()
            },
            halo=meta.halo,
            block_index=meta.block_index,
            blocks_in_sample=meta.blocks_in_sample,
            data=data,
        )


def split_tensor_into_blocks(
    tensor: Tensor,
    block_shape: PerAxis[int],
    *,
    halo: PerAxis[HaloLike],
    stride: Optional[PerAxis[int]] = None,
    pad_mode: PadMode,
) -> Tuple[TotalNumberOfBlocks, Generator[Block, Any, None]]:
    """divide a sample tensor into tensor blocks."""
    n_blocks, block_gen = split_shape_into_blocks(
        tensor.tagged_shape, block_shape=block_shape, halo=halo, stride=stride
    )
    return n_blocks, _block_generator(tensor, block_gen, pad_mode=pad_mode)


def _block_generator(sample: Tensor, blocks: Iterable[BlockMeta], *, pad_mode: PadMode):
    for block in blocks:
        yield Block.from_sample_member(sample, block, pad_mode=pad_mode)



from pathlib import Path
from typing import Any, Tuple, Union

import torch
from torch.jit import ScriptModule

from bioimageio.spec._internal.version_type import Version
from bioimageio.spec.model.v0_5 import ModelDescr, TorchscriptWeightsDescr

from .. import __version__
from ..backends.pytorch_backend import load_torch_model


def convert(
    model_descr: ModelDescr,
    output_path: Path,
    *,
    use_tracing: bool = True,
) -> TorchscriptWeightsDescr:
    """
    Convert model weights from the PyTorch `state_dict` format to TorchScript.

    Args:
        model_descr:
            The model description object that contains the model and its weights in the PyTorch `state_dict` format.
        output_path:
            The file path where the TorchScript model will be saved.
        use_tracing:
            Whether to use tracing or scripting to export the TorchScript format.
            - `True`: Use tracing, which is recommended for models with straightforward control flow.
            - `False`: Use scripting, which is better for models with dynamic control flow (e.g., loops, conditionals).

    Raises:
        ValueError:
            If the provided model does not have weights in the PyTorch `state_dict` format.

    Returns:
        A descriptor object that contains information about the exported TorchScript weights.
    """
    state_dict_weights_descr = model_descr.weights.pytorch_state_dict
    if state_dict_weights_descr is None:
        raise ValueError(
            "The provided model does not have weights in the pytorch state dict format"
        )

    input_data = model_descr.get_input_test_arrays()

    with torch.no_grad():
        input_data = [torch.from_numpy(inp) for inp in input_data]
        model = load_torch_model(state_dict_weights_descr, load_state=True)
        scripted_model: Union[  # pyright: ignore[reportUnknownVariableType]
            ScriptModule, Tuple[Any, ...]
        ] = (
            torch.jit.trace(model, input_data)
            if use_tracing
            else torch.jit.script(model)
        )
        assert not isinstance(scripted_model, tuple), scripted_model

    scripted_model.save(output_path)

    return TorchscriptWeightsDescr(
        source=output_path.absolute(),
        pytorch_version=Version(torch.__version__),
        parent="pytorch_state_dict",
        comment=(
            f"Converted with bioimageio.core {__version__}"
            + f" with use_tracing={use_tracing}."
        ),
    )



from __future__ import annotations

import collections
import warnings
from itertools import product
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    OrderedDict,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import numpy as np
import xarray as xr
from loguru import logger
from numpy.typing import NDArray
from typing_extensions import assert_never

from bioimageio.spec.model.v0_5 import BATCH_AXIS_ID

from .axis import AxisId, PerAxis
from .common import MemberId
from .sample import Sample
from .stat_measures import (
    DatasetMean,
    DatasetMeasure,
    DatasetMeasureBase,
    DatasetPercentile,
    DatasetStd,
    DatasetVar,
    Measure,
    MeasureValue,
    SampleMean,
    SampleMeasure,
    SampleQuantile,
    SampleStd,
    SampleVar,
)
from .tensor import Tensor

try:
    import crick  # pyright: ignore[reportMissingTypeStubs]

except Exception:
    crick = None

    class TDigest:
        def update(self, obj: Any):
            pass

        def quantile(self, q: Any) -> Any:
            pass

else:
    TDigest = crick.TDigest  # type: ignore


class MeanCalculator:
    """to calculate sample and dataset mean for in-memory samples"""

    def __init__(self, member_id: MemberId, axes: Optional[Sequence[AxisId]]):
        super().__init__()
        self._n: int = 0
        self._mean: Optional[Tensor] = None
        self._axes = None if axes is None else tuple(axes)
        self._member_id = member_id
        self._sample_mean = SampleMean(member_id=self._member_id, axes=self._axes)
        self._dataset_mean = DatasetMean(member_id=self._member_id, axes=self._axes)

    def compute(self, sample: Sample) -> Dict[SampleMean, MeasureValue]:
        return {self._sample_mean: self._compute_impl(sample)}

    def _compute_impl(self, sample: Sample) -> Tensor:
        tensor = sample.members[self._member_id].astype("float64", copy=False)
        return tensor.mean(dim=self._axes)

    def update(self, sample: Sample) -> None:
        mean = self._compute_impl(sample)
        self._update_impl(sample.members[self._member_id], mean)

    def compute_and_update(self, sample: Sample) -> Dict[SampleMean, MeasureValue]:
        mean = self._compute_impl(sample)
        self._update_impl(sample.members[self._member_id], mean)
        return {self._sample_mean: mean}

    def _update_impl(self, tensor: Tensor, tensor_mean: Tensor):
        assert tensor_mean.dtype == "float64"
        # reduced voxel count
        n_b = int(tensor.size / tensor_mean.size)

        if self._mean is None:
            assert self._n == 0
            self._n = n_b
            self._mean = tensor_mean
        else:
            assert self._n != 0
            n_a = self._n
            mean_old = self._mean
            self._n = n_a + n_b
            self._mean = (n_a * mean_old + n_b * tensor_mean) / self._n
            assert self._mean.dtype == "float64"

    def finalize(self) -> Dict[DatasetMean, MeasureValue]:
        if self._mean is None:
            return {}
        else:
            return {self._dataset_mean: self._mean}


class MeanVarStdCalculator:
    """to calculate sample and dataset mean, variance or standard deviation"""

    def __init__(self, member_id: MemberId, axes: Optional[Sequence[AxisId]]):
        super().__init__()
        self._axes = None if axes is None else tuple(map(AxisId, axes))
        self._member_id = member_id
        self._n: int = 0
        self._mean: Optional[Tensor] = None
        self._m2: Optional[Tensor] = None

    def compute(
        self, sample: Sample
    ) -> Dict[Union[SampleMean, SampleVar, SampleStd], MeasureValue]:
        tensor = sample.members[self._member_id]
        mean = tensor.mean(dim=self._axes)
        c = (tensor - mean).data
        if self._axes is None:
            n = tensor.size
        else:
            n = int(np.prod([tensor.sizes[d] for d in self._axes]))

        if xr.__version__.startswith("2023"):
            var = (  # pyright: ignore[reportUnknownVariableType]
                xr.dot(c, c, dims=self._axes) / n
            )
        else:
            var = (  # pyright: ignore[reportUnknownVariableType]
                xr.dot(c, c, dim=self._axes) / n
            )

        assert isinstance(var, xr.DataArray)
        std = np.sqrt(var)
        assert isinstance(std, xr.DataArray)
        return {
            SampleMean(axes=self._axes, member_id=self._member_id): mean,
            SampleVar(axes=self._axes, member_id=self._member_id): Tensor.from_xarray(
                var
            ),
            SampleStd(axes=self._axes, member_id=self._member_id): Tensor.from_xarray(
                std
            ),
        }

    def update(self, sample: Sample):
        if self._axes is not None and BATCH_AXIS_ID not in self._axes:
            return

        tensor = sample.members[self._member_id].astype("float64", copy=False)
        mean_b = tensor.mean(dim=self._axes)
        assert mean_b.dtype == "float64"
        # reduced voxel count
        n_b = int(tensor.size / mean_b.size)
        m2_b = ((tensor - mean_b) ** 2).sum(dim=self._axes)
        assert m2_b.dtype == "float64"
        if self._mean is None:
            assert self._m2 is None
            self._n = n_b
            self._mean = mean_b
            self._m2 = m2_b
        else:
            n_a = self._n
            mean_a = self._mean
            m2_a = self._m2
            self._n = n = n_a + n_b
            self._mean = (n_a * mean_a + n_b * mean_b) / n
            assert self._mean.dtype == "float64"
            d = mean_b - mean_a
            self._m2 = m2_a + m2_b + d**2 * n_a * n_b / n
            assert self._m2.dtype == "float64"

    def finalize(
        self,
    ) -> Dict[Union[DatasetMean, DatasetVar, DatasetStd], MeasureValue]:
        if (
            self._axes is not None
            and BATCH_AXIS_ID not in self._axes
            or self._mean is None
        ):
            return {}
        else:
            assert self._m2 is not None
            var = self._m2 / self._n
            sqrt = var**0.5
            if isinstance(sqrt, (int, float)):
                # var and mean are scalar tensors, let's keep it consistent
                sqrt = Tensor.from_xarray(xr.DataArray(sqrt))

            assert isinstance(sqrt, Tensor), type(sqrt)
            return {
                DatasetMean(member_id=self._member_id, axes=self._axes): self._mean,
                DatasetVar(member_id=self._member_id, axes=self._axes): var,
                DatasetStd(member_id=self._member_id, axes=self._axes): sqrt,
            }


class SamplePercentilesCalculator:
    """to calculate sample percentiles"""

    def __init__(
        self,
        member_id: MemberId,
        axes: Optional[Sequence[AxisId]],
        qs: Collection[float],
    ):
        super().__init__()
        assert all(0.0 <= q <= 1.0 for q in qs)
        self._qs = sorted(set(qs))
        self._axes = None if axes is None else tuple(axes)
        self._member_id = member_id

    def compute(self, sample: Sample) -> Dict[SampleQuantile, MeasureValue]:
        tensor = sample.members[self._member_id]
        ps = tensor.quantile(self._qs, dim=self._axes)
        return {
            SampleQuantile(q=q, axes=self._axes, member_id=self._member_id): p
            for q, p in zip(self._qs, ps)
        }


class MeanPercentilesCalculator:
    """to calculate dataset percentiles heuristically by averaging across samples
    **note**: the returned dataset percentiles are an estiamte and **not mathematically correct**
    """

    def __init__(
        self,
        member_id: MemberId,
        axes: Optional[Sequence[AxisId]],
        qs: Collection[float],
    ):
        super().__init__()
        assert all(0.0 <= q <= 1.0 for q in qs)
        self._qs = sorted(set(qs))
        self._axes = None if axes is None else tuple(axes)
        self._member_id = member_id
        self._n: int = 0
        self._estimates: Optional[Tensor] = None

    def update(self, sample: Sample):
        tensor = sample.members[self._member_id]
        sample_estimates = tensor.quantile(self._qs, dim=self._axes).astype(
            "float64", copy=False
        )

        # reduced voxel count
        n = int(tensor.size / np.prod(sample_estimates.shape_tuple[1:]))

        if self._estimates is None:
            assert self._n == 0
            self._estimates = sample_estimates
        else:
            self._estimates = (self._n * self._estimates + n * sample_estimates) / (
                self._n + n
            )
            assert self._estimates.dtype == "float64"

        self._n += n

    def finalize(self) -> Dict[DatasetPercentile, MeasureValue]:
        if self._estimates is None:
            return {}
        else:
            warnings.warn(
                "Computed dataset percentiles naively by averaging percentiles of samples."
            )
            return {
                DatasetPercentile(q=q, axes=self._axes, member_id=self._member_id): e
                for q, e in zip(self._qs, self._estimates)
            }


class CrickPercentilesCalculator:
    """to calculate dataset percentiles with the experimental [crick libray](https://github.com/dask/crick)"""

    def __init__(
        self,
        member_id: MemberId,
        axes: Optional[Sequence[AxisId]],
        qs: Collection[float],
    ):
        warnings.warn(
            "Computing dataset percentiles with experimental 'crick' library."
        )
        super().__init__()
        assert all(0.0 <= q <= 1.0 for q in qs)
        assert axes is None or "_percentiles" not in axes
        self._qs = sorted(set(qs))
        self._axes = None if axes is None else tuple(axes)
        self._member_id = member_id
        self._digest: Optional[List[TDigest]] = None
        self._dims: Optional[Tuple[AxisId, ...]] = None
        self._indices: Optional[Iterator[Tuple[int, ...]]] = None
        self._shape: Optional[Tuple[int, ...]] = None

    def _initialize(self, tensor_sizes: PerAxis[int]):
        assert crick is not None
        out_sizes: OrderedDict[AxisId, int] = collections.OrderedDict(
            _percentiles=len(self._qs)
        )
        if self._axes is not None:
            for d, s in tensor_sizes.items():
                if d not in self._axes:
                    out_sizes[d] = s

        self._dims, self._shape = zip(*out_sizes.items())
        assert self._shape is not None
        d = int(np.prod(self._shape[1:]))
        self._digest = [TDigest() for _ in range(d)]
        self._indices = product(*map(range, self._shape[1:]))

    def update(self, part: Sample):
        tensor = (
            part.members[self._member_id]
            if isinstance(part, Sample)
            else part.members[self._member_id].data
        )
        assert "_percentiles" not in tensor.dims
        if self._digest is None:
            self._initialize(tensor.tagged_shape)

        assert self._digest is not None
        assert self._indices is not None
        assert self._dims is not None
        for i, idx in enumerate(self._indices):
            self._digest[i].update(tensor[dict(zip(self._dims[1:], idx))])

    def finalize(self) -> Dict[DatasetPercentile, MeasureValue]:
        if self._digest is None:
            return {}
        else:
            assert self._dims is not None
            assert self._shape is not None

            vs: NDArray[Any] = np.asarray(
                [[d.quantile(q) for d in self._digest] for q in self._qs]
            ).reshape(self._shape)
            return {
                DatasetPercentile(
                    q=q, axes=self._axes, member_id=self._member_id
                ): Tensor(v, dims=self._dims[1:])
                for q, v in zip(self._qs, vs)
            }


if crick is None:
    DatasetPercentilesCalculator: Type[
        Union[MeanPercentilesCalculator, CrickPercentilesCalculator]
    ] = MeanPercentilesCalculator
else:
    DatasetPercentilesCalculator = CrickPercentilesCalculator


class NaiveSampleMeasureCalculator:
    """wrapper for measures to match interface of other sample measure calculators"""

    def __init__(self, member_id: MemberId, measure: SampleMeasure):
        super().__init__()
        self.tensor_name = member_id
        self.measure = measure

    def compute(self, sample: Sample) -> Dict[SampleMeasure, MeasureValue]:
        return {self.measure: self.measure.compute(sample)}


SampleMeasureCalculator = Union[
    MeanCalculator,
    MeanVarStdCalculator,
    SamplePercentilesCalculator,
    NaiveSampleMeasureCalculator,
]
DatasetMeasureCalculator = Union[
    MeanCalculator, MeanVarStdCalculator, DatasetPercentilesCalculator
]


class StatsCalculator:
    """Estimates dataset statistics and computes sample statistics efficiently"""

    def __init__(
        self,
        measures: Collection[Measure],
        initial_dataset_measures: Optional[
            Mapping[DatasetMeasure, MeasureValue]
        ] = None,
    ):
        super().__init__()
        self.sample_count = 0
        self.sample_calculators, self.dataset_calculators = get_measure_calculators(
            measures
        )
        if not initial_dataset_measures:
            self._current_dataset_measures: Optional[
                Dict[DatasetMeasure, MeasureValue]
            ] = None
        else:
            missing_dataset_meas = {
                m
                for m in measures
                if isinstance(m, DatasetMeasureBase)
                and m not in initial_dataset_measures
            }
            if missing_dataset_meas:
                logger.debug(
                    f"ignoring `initial_dataset_measure` as it is missing {missing_dataset_meas}"
                )
                self._current_dataset_measures = None
            else:
                self._current_dataset_measures = dict(initial_dataset_measures)

    @property
    def has_dataset_measures(self):
        return self._current_dataset_measures is not None

    def update(
        self,
        sample: Union[Sample, Iterable[Sample]],
    ) -> None:
        _ = self._update(sample)

    def finalize(self) -> Dict[DatasetMeasure, MeasureValue]:
        """returns aggregated dataset statistics"""
        if self._current_dataset_measures is None:
            self._current_dataset_measures = {}
            for calc in self.dataset_calculators:
                values = calc.finalize()
                self._current_dataset_measures.update(values.items())

        return self._current_dataset_measures

    def update_and_get_all(
        self,
        sample: Union[Sample, Iterable[Sample]],
    ) -> Dict[Measure, MeasureValue]:
        """Returns sample as well as updated dataset statistics"""
        last_sample = self._update(sample)
        if last_sample is None:
            raise ValueError("`sample` was not a `Sample`, nor did it yield any.")

        return {**self._compute(last_sample), **self.finalize()}

    def skip_update_and_get_all(self, sample: Sample) -> Dict[Measure, MeasureValue]:
        """Returns sample as well as previously computed dataset statistics"""
        return {**self._compute(sample), **self.finalize()}

    def _compute(self, sample: Sample) -> Dict[SampleMeasure, MeasureValue]:
        ret: Dict[SampleMeasure, MeasureValue] = {}
        for calc in self.sample_calculators:
            values = calc.compute(sample)
            ret.update(values.items())

        return ret

    def _update(self, sample: Union[Sample, Iterable[Sample]]) -> Optional[Sample]:
        self.sample_count += 1
        samples = [sample] if isinstance(sample, Sample) else sample
        last_sample = None
        for el in samples:
            last_sample = el
            for calc in self.dataset_calculators:
                calc.update(el)

        self._current_dataset_measures = None
        return last_sample


def get_measure_calculators(
    required_measures: Iterable[Measure],
) -> Tuple[List[SampleMeasureCalculator], List[DatasetMeasureCalculator]]:
    """determines which calculators are needed to compute the required measures efficiently"""

    sample_calculators: List[SampleMeasureCalculator] = []
    dataset_calculators: List[DatasetMeasureCalculator] = []

    # split required measures into groups
    required_sample_means: Set[SampleMean] = set()
    required_dataset_means: Set[DatasetMean] = set()
    required_sample_mean_var_std: Set[Union[SampleMean, SampleVar, SampleStd]] = set()
    required_dataset_mean_var_std: Set[Union[DatasetMean, DatasetVar, DatasetStd]] = (
        set()
    )
    required_sample_percentiles: Dict[
        Tuple[MemberId, Optional[Tuple[AxisId, ...]]], Set[float]
    ] = {}
    required_dataset_percentiles: Dict[
        Tuple[MemberId, Optional[Tuple[AxisId, ...]]], Set[float]
    ] = {}

    for rm in required_measures:
        if isinstance(rm, SampleMean):
            required_sample_means.add(rm)
        elif isinstance(rm, DatasetMean):
            required_dataset_means.add(rm)
        elif isinstance(rm, (SampleVar, SampleStd)):
            required_sample_mean_var_std.update(
                {
                    msv(axes=rm.axes, member_id=rm.member_id)
                    for msv in (SampleMean, SampleStd, SampleVar)
                }
            )
            assert rm in required_sample_mean_var_std
        elif isinstance(rm, (DatasetVar, DatasetStd)):
            required_dataset_mean_var_std.update(
                {
                    msv(axes=rm.axes, member_id=rm.member_id)
                    for msv in (DatasetMean, DatasetStd, DatasetVar)
                }
            )
            assert rm in required_dataset_mean_var_std
        elif isinstance(rm, SampleQuantile):
            required_sample_percentiles.setdefault((rm.member_id, rm.axes), set()).add(
                rm.q
            )
        elif isinstance(rm, DatasetPercentile):
            required_dataset_percentiles.setdefault((rm.member_id, rm.axes), set()).add(
                rm.q
            )
        else:
            assert_never(rm)

    for rm in required_sample_means:
        if rm in required_sample_mean_var_std:
            # computed togehter with var and std
            continue

        sample_calculators.append(MeanCalculator(member_id=rm.member_id, axes=rm.axes))

    for rm in required_sample_mean_var_std:
        sample_calculators.append(
            MeanVarStdCalculator(member_id=rm.member_id, axes=rm.axes)
        )

    for rm in required_dataset_means:
        if rm in required_dataset_mean_var_std:
            # computed togehter with var and std
            continue

        dataset_calculators.append(MeanCalculator(member_id=rm.member_id, axes=rm.axes))

    for rm in required_dataset_mean_var_std:
        dataset_calculators.append(
            MeanVarStdCalculator(member_id=rm.member_id, axes=rm.axes)
        )

    for (tid, axes), qs in required_sample_percentiles.items():
        sample_calculators.append(
            SamplePercentilesCalculator(member_id=tid, axes=axes, qs=qs)
        )

    for (tid, axes), qs in required_dataset_percentiles.items():
        dataset_calculators.append(
            DatasetPercentilesCalculator(member_id=tid, axes=axes, qs=qs)
        )

    return sample_calculators, dataset_calculators


def compute_dataset_measures(
    measures: Iterable[DatasetMeasure], dataset: Iterable[Sample]
) -> Dict[DatasetMeasure, MeasureValue]:
    """compute all dataset `measures` for the given `dataset`"""
    sample_calculators, calculators = get_measure_calculators(measures)
    assert not sample_calculators

    ret: Dict[DatasetMeasure, MeasureValue] = {}

    for sample in dataset:
        for calc in calculators:
            calc.update(sample)

    for calc in calculators:
        ret.update(calc.finalize().items())

    return ret


def compute_sample_measures(
    measures: Iterable[SampleMeasure], sample: Sample
) -> Dict[SampleMeasure, MeasureValue]:
    """compute all sample `measures` for the given `sample`"""
    calculators, dataset_calculators = get_measure_calculators(measures)
    assert not dataset_calculators
    ret: Dict[SampleMeasure, MeasureValue] = {}

    for calc in calculators:
        ret.update(calc.compute(sample).items())

    return ret


def compute_measures(
    measures: Iterable[Measure], dataset: Iterable[Sample]
) -> Dict[Measure, MeasureValue]:
    """compute all `measures` for the given `dataset`
    sample measures are computed for the last sample in `dataset`"""
    sample_calculators, dataset_calculators = get_measure_calculators(measures)
    ret: Dict[Measure, MeasureValue] = {}
    sample = None
    for sample in dataset:
        for calc in dataset_calculators:
            calc.update(sample)
    if sample is None:
        raise ValueError("empty dataset")

    for calc in dataset_calculators:
        ret.update(calc.finalize().items())

    for calc in sample_calculators:
        ret.update(calc.compute(sample).items())

    return ret



from __future__ import annotations

from types import MappingProxyType
from typing import (
    Hashable,
    Literal,
    Mapping,
    NamedTuple,
    Tuple,
    TypeVar,
    Union,
)

from typing_extensions import Self, assert_never

from bioimageio.spec.model import v0_5

from .axis import AxisId

SupportedWeightsFormat = Literal[
    "keras_hdf5",
    "onnx",
    "pytorch_state_dict",
    "tensorflow_saved_model_bundle",
    "torchscript",
]


DTypeStr = Literal[
    "bool",
    "float32",
    "float64",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
]


_LeftRight_T = TypeVar("_LeftRight_T", bound="_LeftRight")
_LeftRightLike = Union[int, Tuple[int, int], _LeftRight_T]


class _LeftRight(NamedTuple):
    left: int
    right: int

    @classmethod
    def create(cls, like: _LeftRightLike[Self]) -> Self:
        if isinstance(like, cls):
            return like
        elif isinstance(like, tuple):
            return cls(*like)
        elif isinstance(like, int):
            return cls(like, like)
        else:
            assert_never(like)


_Where = Literal["left", "right", "left_and_right"]


class CropWidth(_LeftRight):
    pass


CropWidthLike = _LeftRightLike[CropWidth]
CropWhere = _Where


class Halo(_LeftRight):
    pass


HaloLike = _LeftRightLike[Halo]


class OverlapWidth(_LeftRight):
    pass


class PadWidth(_LeftRight):
    pass


PadWidthLike = _LeftRightLike[PadWidth]
PadMode = Literal["edge", "reflect", "symmetric"]
PadWhere = _Where


class SliceInfo(NamedTuple):
    start: int
    stop: int


SampleId = Hashable
"""ID of a sample, see `bioimageio.core.sample.Sample`"""
MemberId = v0_5.TensorId
"""ID of a `Sample` member, see `bioimageio.core.sample.Sample`"""

BlocksizeParameter = Union[
    v0_5.ParameterizedSize_N,
    Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
]
"""
Parameter to determine a concrete size for paramtrized axis sizes defined by
`bioimageio.spec.model.v0_5.ParameterizedSize`.
"""

T = TypeVar("T")
PerMember = Mapping[MemberId, T]

BlockIndex = int
TotalNumberOfBlocks = int


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")

Frozen = MappingProxyType
# class Frozen(Mapping[K, V]):  # adapted from xarray.core.utils.Frozen
#     """Wrapper around an object implementing the mapping interface to make it
#     immutable."""

#     __slots__ = ("mapping",)

#     def __init__(self, mapping: Mapping[K, V]):
#         super().__init__()
#         self.mapping = deepcopy(
#             mapping
#         )  # added deepcopy (compared to xarray.core.utils.Frozen)

#     def __getitem__(self, key: K) -> V:
#         return self.mapping[key]

#     def __iter__(self) -> Iterator[K]:
#         return iter(self.mapping)

#     def __len__(self) -> int:
#         return len(self.mapping)

#     def __contains__(self, key: object) -> bool:
#         return key in self.mapping

#     def __repr__(self) -> str:
#         return f"{type(self).__name__}({self.mapping!r})"



from bioimageio.spec import InvalidDescr, ValidationContext


def test_error_for_wrong_shape(stardist_wrong_shape: str):
    from bioimageio.core._resource_tests import test_model

    summary = test_model(stardist_wrong_shape)
    expected_error_message = (
        "Shape (1, 512, 512, 33) of test output 0 'output' does not match output shape description: "
        "ImplicitOutputShape(reference_tensor='input', "
        "scale=[1.0, 1.0, 1.0, 0.0], offset=[1.0, 1.0, 1.0, 33.0])."
    )
    assert summary.details[0].errors[0].msg == expected_error_message


def test_error_for_wrong_shape2(stardist_wrong_shape2: str):
    from bioimageio.core._resource_tests import test_model

    summary = test_model(stardist_wrong_shape2)
    expected_error_message = (
        "Shape (1, 512, 512, 1) of test input 0 'input' does not match input shape description: "
        "ParameterizedInputShape(min=[1, 80, 80, 1], step=[0, 17, 17, 0])."
    )
    assert summary.details[0].errors[0].msg == expected_error_message


def test_test_model(any_model: str):
    from bioimageio.core._resource_tests import test_model

    with ValidationContext(raise_errors=True):
        summary = test_model(any_model)

    assert summary.status == "passed", summary.display()


def test_loading_description_multiple_times(unet2d_nuclei_broad_model: str):
    from bioimageio.core import load_description

    model_descr = load_description(unet2d_nuclei_broad_model)
    assert not isinstance(model_descr, InvalidDescr)

    # load again, which some users might end up doing
    model_descr = load_description(model_descr)  # pyright: ignore[reportArgumentType]
    assert not isinstance(model_descr, InvalidDescr)



import collections.abc
from pathlib import Path
from typing import (
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from loguru import logger
from tqdm import tqdm

from bioimageio.spec import load_description
from bioimageio.spec.common import PermissiveFileSource
from bioimageio.spec.model import v0_4, v0_5

from ._prediction_pipeline import PredictionPipeline, create_prediction_pipeline
from .axis import AxisId
from .common import BlocksizeParameter, MemberId, PerMember
from .digest_spec import TensorSource, create_sample_for_model, get_member_id
from .io import save_sample
from .sample import Sample


def predict(
    *,
    model: Union[
        PermissiveFileSource, v0_4.ModelDescr, v0_5.ModelDescr, PredictionPipeline
    ],
    inputs: Union[Sample, PerMember[TensorSource], TensorSource],
    sample_id: Hashable = "sample",
    blocksize_parameter: Optional[BlocksizeParameter] = None,
    input_block_shape: Optional[Mapping[MemberId, Mapping[AxisId, int]]] = None,
    skip_preprocessing: bool = False,
    skip_postprocessing: bool = False,
    save_output_path: Optional[Union[Path, str]] = None,
) -> Sample:
    """Run prediction for a single set of input(s) with a bioimage.io model

    Args:
        model: Model to predict with.
            May be given as RDF source, model description or prediction pipeline.
        inputs: the input sample or the named input(s) for this model as a dictionary
        sample_id: the sample id.
            The **sample_id** is used to format **save_output_path**
            and to distinguish sample specific log messages.
        blocksize_parameter: (optional) Tile the input into blocks parametrized by
            **blocksize_parameter** according to any parametrized axis sizes defined
            by the **model**.
            See `bioimageio.spec.model.v0_5.ParameterizedSize` for details.
            Note: For a predetermined, fixed block shape use **input_block_shape**.
        input_block_shape: (optional) Tile the input sample tensors into blocks.
            Note: Use **blocksize_parameter** for a parameterized block shape to
                run prediction independent of the exact block shape.
        skip_preprocessing: Flag to skip the model's preprocessing.
        skip_postprocessing: Flag to skip the model's postprocessing.
        save_output_path: A path with to save the output to. M
            Must contain:
            - `{output_id}` (or `{member_id}`) if the model has multiple output tensors
            May contain:
            - `{sample_id}` to avoid overwriting recurrent calls
    """
    if isinstance(model, PredictionPipeline):
        pp = model
        model = pp.model_description
    else:
        if not isinstance(model, (v0_4.ModelDescr, v0_5.ModelDescr)):
            loaded = load_description(model)
            if not isinstance(loaded, (v0_4.ModelDescr, v0_5.ModelDescr)):
                raise ValueError(f"expected model description, but got {loaded}")
            model = loaded

        pp = create_prediction_pipeline(model)

    if save_output_path is not None:
        if (
            "{output_id}" not in str(save_output_path)
            and "{member_id}" not in str(save_output_path)
            and len(model.outputs) > 1
        ):
            raise ValueError(
                f"Missing `{{output_id}}` in save_output_path={save_output_path} to "
                + "distinguish model outputs "
                + str([get_member_id(d) for d in model.outputs])
            )

    if isinstance(inputs, Sample):
        sample = inputs
    else:
        sample = create_sample_for_model(
            pp.model_description, inputs=inputs, sample_id=sample_id
        )

    if input_block_shape is not None:
        if blocksize_parameter is not None:
            logger.warning(
                "ignoring blocksize_parameter={} in favor of input_block_shape={}",
                blocksize_parameter,
                input_block_shape,
            )

        output = pp.predict_sample_with_fixed_blocking(
            sample,
            input_block_shape=input_block_shape,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )
    elif blocksize_parameter is not None:
        output = pp.predict_sample_with_blocking(
            sample,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
            ns=blocksize_parameter,
        )
    else:
        output = pp.predict_sample_without_blocking(
            sample,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )
    if save_output_path:
        save_sample(save_output_path, output)

    return output


def predict_many(
    *,
    model: Union[
        PermissiveFileSource, v0_4.ModelDescr, v0_5.ModelDescr, PredictionPipeline
    ],
    inputs: Union[Iterable[PerMember[TensorSource]], Iterable[TensorSource]],
    sample_id: str = "sample{i:03}",
    blocksize_parameter: Optional[
        Union[
            v0_5.ParameterizedSize_N,
            Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
        ]
    ] = None,
    skip_preprocessing: bool = False,
    skip_postprocessing: bool = False,
    save_output_path: Optional[Union[Path, str]] = None,
) -> Iterator[Sample]:
    """Run prediction for a multiple sets of inputs with a bioimage.io model

    Args:
        model: Model to predict with.
            May be given as RDF source, model description or prediction pipeline.
        inputs: An iterable of the named input(s) for this model as a dictionary.
        sample_id: The sample id.
            note: `{i}` will be formatted as the i-th sample.
            If `{i}` (or `{i:`) is not present and `inputs` is not an iterable `{i:03}`
            is appended.
        blocksize_parameter: (optional) Tile the input into blocks parametrized by
            blocksize according to any parametrized axis sizes defined in the model RDF.
        skip_preprocessing: Flag to skip the model's preprocessing.
        skip_postprocessing: Flag to skip the model's postprocessing.
        save_output_path: A path to save the output to.
            Must contain:
            - `{sample_id}` to differentiate predicted samples
            - `{output_id}` (or `{member_id}`) if the model has multiple outputs
    """
    if save_output_path is not None and "{sample_id}" not in str(save_output_path):
        raise ValueError(
            f"Missing `{{sample_id}}` in save_output_path={save_output_path}"
            + " to differentiate predicted samples."
        )

    if isinstance(model, PredictionPipeline):
        pp = model
    else:
        if not isinstance(model, (v0_4.ModelDescr, v0_5.ModelDescr)):
            loaded = load_description(model)
            if not isinstance(loaded, (v0_4.ModelDescr, v0_5.ModelDescr)):
                raise ValueError(f"expected model description, but got {loaded}")
            model = loaded

        pp = create_prediction_pipeline(model)

    if not isinstance(inputs, collections.abc.Mapping):
        if "{i}" not in sample_id and "{i:" not in sample_id:
            sample_id += "{i:03}"

        total = len(inputs) if isinstance(inputs, collections.abc.Sized) else None

        for i, ipts in tqdm(enumerate(inputs), total=total):
            yield predict(
                model=pp,
                inputs=ipts,
                sample_id=sample_id.format(i=i),
                blocksize_parameter=blocksize_parameter,
                skip_preprocessing=skip_preprocessing,
                skip_postprocessing=skip_postprocessing,
                save_output_path=save_output_path,
            )
